import os
import csv
from logging import warning

import wrapper
import tsv
from ..itypes import rtypes
from .. import ops
from .. import repops_slice
from ..constants import *

import python
from ..itypes import detector, dimensions,rtypes, dimpaths
from ..utils import nested_array, util


class MatrixTSVRepresentor(tsv.TSVRepresentor):
    def __init__(self, filename, dialect=False, skiprows=-1, col_sections=[], row_sections=[]):
        file = util.open_file(filename)
  
        reader, dialect = self.getReader(file, dialect)
        
        #determine number of rows to skip (e.g. comments)
        if(skiprows == -1):
            skiprows = self.determineSkipRows(file, reader)

        for i in xrange(skiprows):
            file.readline()
        startpos = file.tell()

        #determine type
        colname, colsize = self.convertSection(col_sections, "col_")
        rowname, rowsize = self.convertSection(row_sections, "row_")
                    
        dim1 = dimensions.Dim(UNDEFINED, name="rows")
        dim2 = dimensions.Dim(UNDEFINED, name="cols")
        
        dim_string = dimensions.Dim(UNDEFINED)
        bytes = rtypes.TypeBytes(dims=dimpaths.DimPath(dim_string))

        types = []
        names = []
        for coln,cols in zip(colname, colsize):
            if cols == 1:
                t =rtypes.TypeArray(dims=dimpaths.DimPath(dim1),subtypes=(bytes,))
            else:
                dim = dimensions.Dim(cols, name=coln+"_dim")
                subtype = rtypes.TypeArray(dims=dimpaths.DimPath(dim),subtypes=(bytes,))            
                t = rtypes.TypeArray(dims=dimpaths.DimPath(dim1),subtypes=(subtype,))
            types.append(t)
            names.append(coln)
        for rown,rows in zip(rowname, rowsize):
            if rows == 1:
                t =rtypes.TypeArray(dims=dimpaths.DimPath(dim2),subtypes=(bytes,))
            else:
                dim = dimensions.Dim(rows, name=rown+"_dim")
                subtype = rtypes.TypeArray(dims=dimpaths.DimPath(dim2),subtypes=(bytes,))            
                t = rtypes.TypeArray(dims=dimpaths.DimPath(dim),subtypes=(subtype,))
            types.append(t)
            names.append(rown)
        subtype = rtypes.TypeArray(dims=dimpaths.DimPath(dim2),subtypes=(bytes,))            
        types.append(rtypes.TypeArray(dims=dimpaths.DimPath(dim1),subtypes=(subtype,)))
        names.append('matrix')

        dtype = rtypes.TypeTuple(subtypes=tuple(types), fieldnames=tuple(names))
        slice = MatrixTSVOp(filename, dialect, startpos, dtype, colsize, rowsize, "data")
        nslices = repops_slice.UnpackTuple._apply(slice)
        file.close()
        self._initialize(tuple(nslices))
    
    def convertSection(self, section, prefix):
        colsize = []
        colname = []
        for pos,c in enumerate(section):
            if isinstance(c,basestring):
                colsize.append(1)
                colname.append(c)
            elif isinstance(c,dict):
                assert len(c) == 1, "Only one element in col_section dicts allowed"
                colname.append(c.keys()[0])
                colsize.append(c.values()[0])
            elif isinstance(c, int):
                assert c >= 1, "Only matrix section sizes >= 1 allowed"
                colsize.append(c)
                if pos < 26:
                    colname.append(prefix + chr(pos + 97))
                else:
                    colname.append(prefix + str(pos))
        colname = [util.valid_name(coln) for coln in colname]

        return (colname, colsize)

class MatrixTSVOp(ops.ExtendOp):
    __slots__ = ["filename", "dialect","startpos","colsize","rowsize"]

    def __init__(self, filename, dialect, startpos, rtype, colsize, rowsize, name):
        self.filename = filename
        self.dialect = dialect
        self.startpos = startpos
        self.rowsize = rowsize
        self.colsize = colsize
        ops.ExtendOp.__init__(self,name=name,rtype=rtype)

    def py_exec(self):
        file = util.open_file(self.filename)
        file.seek(self.startpos)

        if(issubclass(self.dialect, csv.Dialect)):
            reader = csv.reader(file, self.dialect)
        else:
            reader = csv.reader(file, delimiter=self.dialect)
        data = [tuple(row) for row in reader]
        
        xlen = len(data[0])
        for pos, row in enumerate(data):
            if len(row) != xlen:
                warning('Removing last %d rows', len(data) - pos)            
                data = data[:pos]


        file.close()
        
        cols = []
        rows = []
        skiprows = sum(self.rowsize)
        for cs in self.colsize:
            ndata = []
            ncol = []
            if cs == 1:
                for row in data:
                    ndata.append(row[1:])
                    ncol.append(row[0])
            else:
                for row in data:
                    ndata.append(row[cs:])
                    ncol.append(row[:cs])
            cols.append(ncol[skiprows:])
            data = ndata
        
        for rs in self.rowsize:
            if rs == 1:
                nrow = data[0]
                data = data[1:]
            else:
                nrow = data[:rs]
                data = data[rs:]
            rows.append(nrow)
        
        data = tuple(cols) + tuple(rows) + (data,)
        ndata = nested_array.NestedArray(data,self.type)
        return python.ResultOp.from_slice(ndata,self)




import operator
import repops
from constants import *
from itypes import rtypes
import ops
_delay_import_(globals(),"utils","util")
_delay_import_(globals(),"representor")
_delay_import_(globals(),"wrappers","python")
_delay_import_(globals(),"itypes", "casts","dimpaths","dimensions")

class FuncSignature(object):
    def __init__(self, signame):
        self.name = signame
    def check(self, **kwargs):
        return True

class Param(object):
    __slots__ = ["name", "type","default"]
    def __init__(self,name, type=None, default=NOVAL):
        assert isinstance(name, basestring), "Param name should be a string"
        assert type is None or isinstance(type,rtypes.TypeUnknown), "Type should be None or a type from module rtypes"

        self.name = name
        self.type = type
        self.default = default

    def __str__(self):
       if(self.type is None):
           res = self.name
       else:
           res = str(self.type) + " " + self.name
       if(not self.default is NOVAL):
           if(isinstance(self.default,representor.Representor)):
               res += "=" + str(self.default())
           else:
               res += "=" + str(self.default)
       return res

    def check(self,val):
        if(val is NOVAL):
            if(self.default is NOVAL):
                return False
            else:
                return self.default

        if(self.type is None):
            return val
        else:
            if(not isinstance(val, ops.UnaryOp)):
                return False
            if not isinstance(val.type, self.type.__class__):
                return False
        return val

    def withNumber(self, number):
        return self.__class__(self.name + str(number), self.type, self.default)


class ParamFuncSignature(object):
    __slots__ = ['signame', 'inparams', 'outparam']

    def __init__(self, inparams, outparams, signame=None):
        assert all(isinstance(p, Param) for p in inparams), "Inparams should be of param type"
        assert isinstance(outparam, Param), "Outparamsshould be of param type"
        self.inparams = inparams
        self.outparam = outparam
        
        if(signame is None):
            inparamstr = ",".join(map(str,inparams))
            outparamstr = str(outparam)
            signame = inparamstr + "-->" + outparamstr
        FuncSignature.__init__(signame)
            
    def check(self, **kwargs):
        res = {}
        for param in self.inparams:
            pc = param.check(kwargs.get(param.name,None))
            if(pc is False):
                return False
            res[param.name] = pc
        return (res, self.outparams)

class Func(object):
    _sigs = []

    @classmethod
    def _findSignature(cls, **kwargs):
        for sig in cls._sigs:
            res = sig.check(**kwargs)
            if(res is False):
                continue
            if(isinstance(res,tuple)):
                nkwargs, outdescr = res
            else:
                nkwargs = kwargs.copy()
                outdescr = res
            return (sig, nkwargs, outdescr)
        
        #cannot find anything
        res = []
        for field,slice in kwargs.iteritems():
            if(isinstance(slice, ops.UnaryOp)):
                res.append((str(slice.type) + " ",field))
            else:
                res.append(("", field + "=" + str(slice)))
        res.sort(key=operator.itemgetter(1))
        res = [a + b for a,b in res]
        raise RuntimeError, "Cannot find func " + str(cls.__name__)  + " with signature (" + ", ".join(res) + ")"
        
       

class UnaryFuncOp(repops.UnaryOpRep, Func):
    def __init__(self, source, *params, **kwargs):
        if(not isinstance(source, representor.Representor)):
            source = python.Rep(source,name="data")
        repops.UnaryOpRep.__init__(self,source, *params,**kwargs)

class UnaryFuncElemOp(UnaryFuncOp):
    def _process(self, source, **kwargs):
        if not source._typesKnown():
            return
        nslices = []
        for slice in source._slices:
            slice = self._prepareSlice(slice)
            sig, nkwargs, outparam = self._findSignature(slice=slice,**kwargs)
            s = ops.UnaryFuncElemOp(self.__class__.__name__, sig, outparam, **nkwargs)
            s = self._finishSlice(s)
            nslices.append(s)
        return self._initialize(tuple(nslices))

    def _prepareSlice(self,slice):
        return slice

    def _finishSlice(self, slice):
        return slice

class UnaryFuncDimOp(UnaryFuncOp):
    _slicecls = ops.UnaryFuncSeqOp

    def _process(self, source, dim=None, **kwargs):
        if not source._typesKnown():
            return
        nslices = self._apply(source._slices, dim, **kwargs)
        return self._initialize(tuple(nslices))

    @classmethod
    def _apply(cls, xslices, dim,  **kwargs):
        nslices = []
        found = False
        for slice in xslices:
            selpath = dimpaths.identifyUniqueDimPath([slice.dims],dim)
            lastposs = slice.dims.matchDimPath(selpath)
            if(lastposs):
                slice = cls._prepareSlice(slice)
                found = True
            for lastpos in lastposs:
                packdepth = len(slice.dims) - lastpos
                sig, nkwargs, outparam = cls._findSignature(slice=slice, packdepth=packdepth, **kwargs)
                slice = cls._slicecls(cls.__name__, sig, outparam, **nkwargs)
            if(lastposs):
                slice = cls._finishSlice(slice)
            nslices.append(slice)

        if(not found):
            raise RuntimeError, "No slice with dims to apply " + cls.__name__
        
        return nslices

    @classmethod 
    def _prepareSlice(self,slice):
        return slice
    @classmethod
    def _finishSlice(self, slice):
        return slice

class UnaryFuncAggregateOp(UnaryFuncDimOp):
    _slicecls = ops.UnaryFuncAggregateOp

class BinaryFuncOp(repops.MultiOpRep, Func):
    def __init__(self, lsource, rsource, **kwargs):
        if(not isinstance(lsource, representor.Representor)):
            lsource = repops.PlusPrefix(python.Rep(lsource,name="data"))
        if(not isinstance(rsource, representor.Representor)):
            rsource = repops.PlusPrefix(python.Rep(rsource,name="data"))
        repops.MultiOpRep.__init__(self,(lsource,rsource), **kwargs)

        
class BinaryFuncElemOp(BinaryFuncOp):
    _allow_partial_bc = False
    def _process(self, sources, **kwargs):
        lsource,rsource = sources
        if not lsource._typesKnown() or not rsource._typesKnown():
            return

        if(isinstance(lsource,repops.PlusPrefix) or isinstance(rsource,repops.PlusPrefix)):
            mode = "pos"
        else:
            mode = "dim"
        
        nslices = []
        nslice = max(len(lsource._slices), len(rsource._slices))
        for pos, binslices in enumerate(util.zip_broadcast(lsource._slices, rsource._slices)):
            if(nslice <= 1):
                pos = None
            nslices.append(self._apply(binslices, mode, pos, **kwargs))

        return self._initialize(tuple(nslices))

    @classmethod
    def _apply(cls, binslices, mode, pos=None, **kwargs):
        binslices = cls._prepareSlices(*binslices)
        (lslice,rslice),plans = ops.broadcast(binslices,mode, partial=cls._allow_partial_bc)
        sig, nkwargs, outparam = cls._findSignature(left=lslice,right=rslice,**kwargs)
        if(isinstance(outparam, rtypes.TypeUnknown)):
            if(binslices[0].name == binslices[1].name):
                name = binslices[0].name
            elif(binslices[0].name == "data"):
                name = binslices[1].name
            elif(binslices[1].name == "data"):
                name = binslices[0].name
            else:
                name = "result"
            outparam = Param(name, outparam)
            if(not pos is None and name == "result"):
                outparam = outparam.withNumber(pos)
        s = ops.BinFuncElemOp(cls.__name__, sig,outparam, **nkwargs)
        return s

    @classmethod
    def _prepareSlices(cls, lslice, rslice):
        return (lslice,rslice)


class WithinSignature(FuncSignature):
    def check(self, left, right):#{{{
        in1_type = left.type
        in2_type = right.type

        out_cls = rtypes.TypeBool
        out_type = out_cls(in1_type.has_missing or in2_type.has_missing)
        return out_type#}}}

within_sig = WithinSignature("within")

class Within(BinaryFuncElemOp):
    _allow_partial_bc = True
    _sigs = [within_sig]
    @classmethod
    def _prepareSlices(self,lslice,rslice):
        return (lslice, ops.PackArrayOp(ops.ensure_frozen(rslice),1))

class Contains(BinaryFuncElemOp):
    _allow_partial_bc = True
    _sigs = [within_sig]
    @classmethod
    def _prepareSlices(self,lslice,rslice):
        return (ops.PackArrayOp(ops.ensure_frozen(lslice),1), rslice)

class BinArithSignature(FuncSignature):
    def check(self, left, right):
        in1_type = left.type
        in2_type = right.type
        if(not isinstance(in1_type, rtypes.TypeNumber) or not isinstance(in2_type, rtypes.TypeNumber)):
            return False

        in1_impli_cls = casts.findImplicitCastTypes(in1_type.__class__)
        in2_impli_cls = casts.findImplicitCastTypes(in2_type.__class__)
        out_impli_cls = in1_impli_cls & in2_impli_cls
        
        out_impli_cls = [ocls for ocls in out_impli_cls \
                                if issubclass(ocls,rtypes.TypeNumber)]
        if(not out_impli_cls):
            return False
    
        out_impli_cls = rtypes.mostSpecializedTypesCls(out_impli_cls)
        assert len(out_impli_cls) == 1, "Multiple output types for " + \
                    "arithmetic operation found: " + str(out_impli_cls)
        out_cls = out_impli_cls[0]
        
        out_type = out_cls(in1_type.has_missing or in2_type.has_missing)

        return out_type#}}}
bin_arithsig = BinArithSignature("simple_arith")

class SetSetSignature(FuncSignature):
    def check(self, left, right):
        in1_type = left.type
        in2_type = right.type
        if(not isinstance(left.type, rtypes.TypeSet) or not isinstance(right.type, rtypes.TypeSet)):
            return False
        
        subout_type = casts.castImplicitCommonType(left.type.subtypes[0],right.type.subtypes[0])
        
        ldim = left.type.dims[0]
        rdim = right.type.dims[0]
        dim = dimensions.Dim(UNDEFINED,(True,) * len(left.dims),ldim.has_missing or rdim.has_missing, name=ldim.name + "_" + rdim.name)
        dims = dimpaths.DimPath(dim)

        out_type = rtypes.TypeSet(has_missing=left.type.has_missing or right.type.has_missing, subtypes=(subout_type,), dims=dims)
        return out_type#}}}
setset_sig = SetSetSignature("simple_arith")

class StringAddStringSignature(FuncSignature):
    def check(self, left, right):
        in1_type = left.type
        in2_type = right.type
        if(not isinstance(in1_type, rtypes.TypeString) or not isinstance(in2_type, rtypes.TypeString)):
            return False
       
        if(in1_type.dims[0].shape == UNDEFINED or in2_type.dims[0].shape == UNDEFINED):
            nshape =  UNDEFINED
        else:
            nshape = in1_type.dims[0].shape + in2_type.dims[0].shape

        in1_impli_cls = casts.findImplicitCastTypes(in1_type.__class__)
        in2_impli_cls = casts.findImplicitCastTypes(in2_type.__class__)
        out_impli_cls = in1_impli_cls & in2_impli_cls
        out_impli_cls = rtypes.mostSpecializedTypesCls(out_impli_cls)
        assert len(out_impli_cls) == 1, "Multiple output types for " + \
                    "arithmetic operation found: " + str(out_impli_cls)
        out_cls = out_impli_cls[0]
        
        ndim = dimensions.Dim(nshape,dependent=(True,) * len(left.dims), name = in1_type.dims[0].name + "_" + in2_type.dims[0].name)
        out_type = out_cls(in1_type.has_missing or in2_type.has_missing, dims=dimpaths.DimPath(ndim))
        return out_type#}}}
stringaddstring_sig = StringAddStringSignature("string_add_string")

class ArrayAddArraySignature(FuncSignature):
    def check(self, left, right):
        in1_type = left.type
        in2_type = right.type
        if(not in1_type.__class__ is rtypes.TypeArray or not in2_type.__class__ is rtypes.TypeArray):
            return False

        if(in1_type.dims[1:]):
            subtype1 = rtypes.TypeArray(subtypes=in1_type.subtypes, dims=in1_type.dims[1:])
        else:
            subtype1 = in1_type.subtypes[0]
        
        if(in2_type.dims[1:]):
            subtype2 = rtypes.TypeArray(subtypes=in2_type.subtypes, dims=in2_type.dims[1:])
        else:
            subtype2 = in2_type.subtypes[0]

        nsubtype = casts.castImplicitCommonType(subtype1,subtype2)
        if(nsubtype is False):
            return False
            
        if(in1_type.dims[0].shape == UNDEFINED or in2_type.dims[0].shape == UNDEFINED):
            nshape =  UNDEFINED
        else:
            nshape = in1_type.dims[0].shape + in2_type.dims[0].shape
        
        ndim = dimensions.Dim(nshape,dependent=(True,) * len(left.dims), name = in1_type.dims[0].name + "_" + in2_type.dims[0].name)
        out_type = rtypes.TypeArray(in1_type.has_missing or in2_type.has_missing, dims=dimpaths.DimPath(ndim), subtypes=(nsubtype,))
        return out_type#}}}
arrayaddarray_sig = ArrayAddArraySignature("array_add_array")

class Add(BinaryFuncElemOp):
   _sigs = [bin_arithsig, stringaddstring_sig, arrayaddarray_sig]

class Subtract(BinaryFuncElemOp):
   _sigs = [bin_arithsig, setset_sig]

class Multiply(BinaryFuncElemOp):
   _sigs = [bin_arithsig]

class Modulo(BinaryFuncElemOp):
   _sigs = [bin_arithsig]

class Divide(BinaryFuncElemOp):
   _sigs = [bin_arithsig]

class FloorDivide(BinaryFuncElemOp):
   _sigs = [bin_arithsig]

class BoolOutSignature(FuncSignature):
    def __init__(self,name,clss):
        self.comparecls = clss
        FuncSignature.__init__(self,name)

    def check(self, left, right):#{{{
        in1_type = left.type
        in2_type = right.type

        if(not isinstance(in1_type, self.comparecls) or not isinstance(in2_type, self.comparecls)):
            return False

        out_cls = rtypes.TypeBool
        out_type = out_cls(in1_type.has_missing or in2_type.has_missing)

        return out_type#}}}
       
boolboolsig = BoolOutSignature("boolbool",rtypes.TypeBool)

class And(BinaryFuncElemOp):
   _sigs = [boolboolsig, bin_arithsig, setset_sig]

class Or(BinaryFuncElemOp):
   _sigs = [boolboolsig, bin_arithsig, setset_sig]

class Xor(BinaryFuncElemOp):
   _sigs = [boolboolsig, bin_arithsig, setset_sig]

class Power(BinaryFuncElemOp):
   _sigs = [bin_arithsig]


       
compareanysig = BoolOutSignature("simple_cmp",rtypes.TypeAny)
comparesetsig = BoolOutSignature("simple_cmp",rtypes.TypeSet)
comparestringsig = BoolOutSignature("string_cmp",rtypes.TypeString)

class Equal(BinaryFuncElemOp):
    _sigs = [comparestringsig,comparesetsig,compareanysig]

class NotEqual(BinaryFuncElemOp):
    _sigs = [comparestringsig,comparesetsig,compareanysig]

class LessEqual(BinaryFuncElemOp):
    _sigs = [comparestringsig,comparesetsig,compareanysig]

class Less(BinaryFuncElemOp):
    _sigs = [comparestringsig,comparesetsig,compareanysig]
    
class GreaterEqual(BinaryFuncElemOp):
    _sigs = [comparestringsig,comparesetsig,compareanysig]

class Greater(BinaryFuncElemOp):
    _sigs = [comparestringsig,comparesetsig,compareanysig]

class MergeSignature(FuncSignature):
    def check(self, left, right):
        in1_type = left.type
        in2_type = right.type

        res = casts.castImplicitCommonType(in1_type, in2_type)
        if res is False:
            return False
        return res#}}}
mergesig = MergeSignature("merge")

class Merge(BinaryFuncElemOp):
    _sigs = [mergesig]

class EachSignature(FuncSignature):
    def check(self, slice, eachfunc, dtype=rtypes.unknown):#{{{
        if(not isinstance(dtype,rtypes.TypeUnknown)):
            dtype = rtypes.createType(dtype,len(slice.dims)) 
        nkwargs = {'eachfunc': eachfunc, 'slice':slice}
        return (nkwargs, Param(slice.name, dtype))#}}}
eachsig = EachSignature("each")

class Each(UnaryFuncElemOp):
    _sigs = [eachsig]


class UnaryArithSignature(FuncSignature):
    def check(self, slice):#{{{
        in_type = slice.type
        if(not isinstance(in_type, rtypes.TypeNumber)):
            return False

        return Param(slice.name, in_type)#}}}

unary_arithsig = UnaryArithSignature("number")

class UnaryBoolSignature(FuncSignature):
    def check(self, slice):#{{{
        in_type = slice.type
        if(not in_type.has_missing or not isinstance(in_type, rtypes.TypeBool)):
            return False

        return Param(slice.name, in_type)#}}}

unary_boolsig = UnaryBoolSignature("bool")

class Invert(UnaryFuncElemOp):
    _sigs = [unary_boolsig, unary_arithsig]

class Abs(UnaryFuncElemOp):
    _sigs = [unary_arithsig]

class Negative(UnaryFuncElemOp):
    _sigs = [unary_arithsig]


class ReplaceMissingSig(FuncSignature):
    def check(self, slice, def_value=NOVAL):#{{{
        in_type = slice.type
        nin_type = in_type.copy()
        #FIXME: adapt dim in case of array
        nin_type = nin_type.setHasMissing(False)

        return Param(slice.name, nin_type)#}}}

repmissing = ReplaceMissingSig("repmissing")

class ReplaceMissing(UnaryFuncElemOp):
    _sigs =[repmissing]


class IsMissingSig(FuncSignature):
    def check(self, slice, def_value=NOVAL):#{{{
        o_type = rtypes.TypeBool()
        return Param(slice.name, o_type)#}}}
ismissingsig = IsMissingSig("ismissing")

class IsMissing(UnaryFuncElemOp):
    _sigs =[ismissingsig]

class UnaryFixShapeSignature(FuncSignature):
    def __init__(self,name,check_dependent=True):
        self.check_dependent = check_dependent
        FuncSignature.__init__(self,name)

    def check(self,slice,packdepth):
        if(self.check_dependent):
            for pos,i in enumerate(range(len(slice.dims) - packdepth + 1, len(slice.dims))):
                d = slice.dims[i]
                if len(d.dependent) > pos and d.dependent[pos] == True:
                    return False
        return True

class UnaryTypeToTypeSignature(UnaryFixShapeSignature):
    def __init__(self, name, itypecls, otypecls=None, check_dependent=True):
        self.itypecls = itypecls
        self.otypecls = otypecls
        UnaryFixShapeSignature.__init__(self, name, check_dependent)

    def check(self, slice, packdepth, **kwargs):#{{{
        if not UnaryFixShapeSignature.check(self,slice,packdepth):
            return False

        in_type = slice.type
        if(not isinstance(in_type, self.itypecls)):
            return False

        if(self.otypecls is None):
            nstype = in_type
        else:
            nstype = self.otypecls(in_type.has_missing)
        return Param(slice.name, nstype)#}}}

any_nodepsig = UnaryTypeToTypeSignature("any_nodep", rtypes.TypeAny, rtypes.TypePlatformInt, check_dependent=False)

boolsig = UnaryTypeToTypeSignature("fixdim", rtypes.TypeBool, rtypes.TypeBool)
numbersig = UnaryTypeToTypeSignature("fixdim", rtypes.TypeNumber)
int_tointsig = UnaryTypeToTypeSignature("fixdim", rtypes.TypeInteger,rtypes.TypePlatformInt)
float_tofloatsig = UnaryTypeToTypeSignature("fixdim", rtypes.TypeReal64,rtypes.TypeReal64)
number_tofloatsig = UnaryTypeToTypeSignature("fixdim", rtypes.TypeNumber,rtypes.TypeReal64)
number_tointsig = UnaryTypeToTypeSignature("fixdim", rtypes.TypeNumber,rtypes.TypePlatformInt)

class UnarySortableSignature(UnaryFixShapeSignature):
    def __init__(self, name, check_dependent=True):
        UnaryFixShapeSignature.__init__(self, name, check_dependent)

    def check(self, slice, packdepth, **kwargs):#{{{
        if not UnaryFixShapeSignature.check(self,slice,packdepth):
            return False

        in_type = slice.type
        if(in_type.__class__ == rtypes.TypeArray):
            return False

        nstype = rtypes.TypePlatformInt()
        return Param(slice.name, nstype)#}}}

sortablesig = UnarySortableSignature("sortable")
class Argsort(UnaryFuncDimOp):
    _sigs = [sortablesig]

sortablesig = UnarySortableSignature("sortable")
class Rank(UnaryFuncDimOp):
    _sigs = [sortablesig]

class Pos(UnaryFuncDimOp):
    _sigs = [any_nodepsig]

class CumSum(UnaryFuncDimOp):
    _sigs = [int_tointsig, float_tofloatsig]

class Any(UnaryFuncAggregateOp):
    _sigs = [boolsig]

class All(UnaryFuncAggregateOp):
    _sigs = [boolsig]

class Max(UnaryFuncAggregateOp):
    _sigs = [numbersig]

class Min(UnaryFuncAggregateOp):
    _sigs = [numbersig]

class UnaryConcatenateSignature(UnaryFixShapeSignature):
    def check(self, slice, packdepth, **kwargs):#{{{
        if not UnaryFixShapeSignature.check(self,slice,packdepth):
            return False
        in_type = slice.type
        if(not in_type.__class__ is rtypes.TypeArray):
            return False
        if(not in_type.dims):
            return False

        curdim = in_type.dims[0]
        adim = slice.dims[-packdepth]
        
        if(not curdim.shape is UNDEFINED and not adim.shape is UNDEFINED):
            nshape = curdim.shape * adim.shape
        else:
            nshape = UNDEFINED
        nname = "s" + curdim.name

        #note: in_type.dims[0].dependent will be updated by aggregate slice
        ndim = dimensions.Dim(nshape, in_type.dims[0].dependent, has_missing=in_type.dims[0].has_missing, name = nname)
        nin_type = in_type._updateDepDim(0,ndim)

        return Param(slice.name, nin_type)#}}}

concatenate_sig = UnaryConcatenateSignature("arrayarray")

class UnaryStringConcatenateSignature(UnaryFixShapeSignature):
    def check(self, slice, packdepth, **kwargs):#{{{
        if not UnaryFixShapeSignature.check(self,slice,packdepth):
            return False
        in_type = slice.type
        if(not isinstance(in_type,rtypes.TypeString)):
            return False
        if(not in_type.dims):
            return False

        curdim = in_type.dims[0]
        adim = slice.dims[-packdepth]
        
        if(not curdim.shape is UNDEFINED and not adim.shape is UNDEFINED):
            nshape = curdim.shape * adim.shape
        else:
            nshape = UNDEFINED
        nname = "s" + curdim.name

        #note: in_type.dims[0].dependent will be updated by aggregate slice
        ndim = dimensions.Dim(nshape, in_type.dims[0].dependent, has_missing=in_type.dims[0].has_missing, name = nname)
        nin_type = in_type._updateDepDim(0,ndim)

        return Param(slice.name, nin_type)#}}}
strconcatenate_sig = UnaryStringConcatenateSignature("stringstring")

class Sum(UnaryFuncAggregateOp):
    _sigs = [int_tointsig, float_tofloatsig, concatenate_sig, strconcatenate_sig]

class Mean(UnaryFuncAggregateOp):
    _sigs = [number_tofloatsig]

class Argmax(UnaryFuncAggregateOp):
    _sigs = [number_tointsig]

class Argmin(UnaryFuncAggregateOp):
    _sigs = [number_tointsig]

class Median(UnaryFuncAggregateOp):
    _sigs = [number_tofloatsig]

class Std(UnaryFuncAggregateOp):
    _sigs = [number_tofloatsig]

class CountSignature(FuncSignature):
    def check(self, slice, packdepth):#{{{
        if(packdepth > 1):
            return False
        
        nstype = rtypes.TypePlatformInt(len(slice.dims) > 1 and slice.dims[-2].has_missing)
        return Param(slice.name, nstype)#}}}
countsig = CountSignature("count")

class Count(UnaryFuncAggregateOp):
    _sigs = [countsig]

class SetSignature(UnaryFixShapeSignature):
    def check(self, slice, packdepth):#{{{
        if not UnaryFixShapeSignature.check(self,slice,packdepth):
            return False
        
        has_missing = slice.dims[-1].has_missing
        subtypes = (slice.type,)
        dim = dimensions.Dim(UNDEFINED,(True,) * len(slice.dims), slice.type.has_missing, name="s" + slice.dims[-packdepth].name)
        nstype = rtypes.TypeSet(has_missing,dimpaths.DimPath(dim),subtypes)
        return Param(slice.name, nstype)#}}}
setsig = SetSignature("set")

class Set(UnaryFuncAggregateOp):
    _sigs = [setsig]

    @classmethod
    def _prepareSlice(self,slice):
        return ops.ensure_frozen(slice)

class UniqueSignature(UnaryFixShapeSignature):
    def check(self, slice, packdepth):#{{{
        if not UnaryFixShapeSignature.check(self,slice,packdepth):
            return False
        
        has_missing = slice.dims[-1].has_missing
        subtypes = (rtypes.TypePlatformInt(),)
        dim = dimensions.Dim(UNDEFINED,(True,) * len(slice.dims), slice.type.has_missing, name="s" + slice.dims[-packdepth].name)
        nstype = rtypes.TypeArray(has_missing,dimpaths.DimPath(dim),subtypes)
        return Param(slice.name, nstype)#}}}
uniquesig = UniqueSignature("unique")

class Argunique(UnaryFuncAggregateOp):
    _sigs = [uniquesig]

    @classmethod
    def _prepareSlice(self,slice):
        return ops.ensure_frozen(slice)

class CorrSignature(FuncSignature):
    def check(self, slice):
        if(len(slice.dims) < 2):
            return False

        if(not isinstance(slice.type,rtypes.TypeNumber)):
            return False

        corrdim = slice.dims[-2]
        corrdim2 = corrdim.insertDepDim(0,corrdim)
        ndims = dimpaths.DimPath(corrdim, corrdim2)

        ntype = rtypes.TypeReal64(slice.type.has_missing)
        ntype = dimpaths.dimsToArrays(ndims, ntype)

        slice = ops.PackArrayOp(slice,2)
        nkwargs = {'slice':slice}
        return (nkwargs, Param(slice.name, ntype))
corrsig = CorrSignature("corr")

class Corr(UnaryFuncElemOp):
    _sigs = [corrsig]

    def _finishSlice(self,slice):
        return ops.UnpackArrayOp(slice,2)

        




UNDEFINED = -1


BCCOPY=0       #exists here, exists there
BCENSURE=1     #exists here, exists there, but has different dim identity

BCSOURCE=2     #exists here, not there

BCNEW=3        #exists not here, needs broadcasting
BCEXIST=4      #exists not here, but broadcast dim does exist


NOVAL="NOVALUE" #alternative None for functions which also accept NOne as valid input
                #but still want some params to be optional


#slice matching
COMMON_POS="COMMONPOS"
COMMON_NAME="COMMONNAME"


LASTCOMMONDIM = -1L
LCDIM = LASTCOMMONDIM

class NewDim(object):
    def __init__(self,name=None):
        self.name = name
    def __call__(self,name=None):
        return NewDim(name)
newdim = NewDim()
NEWDIM = newdim






import manager
import create_graph
import annotate_replinks
from ..utils import util
from ..utils.multi_visitor import VisitorFactory, NF_ERROR, F_CACHE
from ..constants import *
_delay_import_(globals(),"..query_graph")
_delay_import_(globals(),"..ops")

class PrePeepHole(VisitorFactory(prefixes=("visit",), flags=NF_ERROR), manager.Pass):
    after=set([annotate_replinks.AnnotateRepLinks])

    @classmethod
    def run(cls, query, run_manager):
        self = cls()
        self.graph = run_manager.pass_results[create_graph.CreateGraph]

        for node in list(self.graph.nodes):
            self.visit(node)
    
    def visitNode(self, node):
        pass

    def remove_unaryop(self, node):
        source = self.graph.getDataEdge(node).source
        target_edges = list(self.graph.edge_source[node])
        self.graph.dropNode(node)
        for target_edge in target_edges:
            assert isinstance(target_edge,query_graph.ParamEdge), "Unknown edge type encountered"
            target_edge.source = source
            self.graph.addEdge(target_edge)
    visitNoOp=remove_unaryop
    visitChangeBookmarkOp=remove_unaryop
    visitChangeNameOp=remove_unaryop
    visitChangeDimPathOp=remove_unaryop

    def visitPackArrayOp(self,node):
        target_edges = list(self.graph.edge_source[node])
        if(len(target_edges) != 1): 
            return

        target = target_edges[0].target
        if(not isinstance(target, ops.UnpackArrayOp)):
            return
        
        pack_depth = len(node.pack_dims)
        unpack_depth = len(target.unpack_dims)
        return self.combine_pack_unpack(pack_depth, unpack_depth, node, target)

    def visitUnpackArrayOp(self,node):
        target_edges = list(self.graph.edge_source[node])
        if(len(target_edges) != 1): 
            return

        target = target_edges[0].target
        if(not isinstance(target, ops.PackArrayOp)):
            return
        
        unpack_depth = len(node.unpack_dims)
        pack_depth = len(target.pack_dims)
        return self.combine_pack_unpack(pack_depth, unpack_depth, node, target)

    def combine_pack_unpack(self, pack_depth, unpack_depth, node, target):
        sourceedge = self.graph.getDataEdge(node)
        self.graph.dropEdge(sourceedge)
        if(pack_depth != unpack_depth): 
            if(pack_depth > unpack_depth):
                nsource = ops.PackArrayOp(sourceedge.source,pack_depth - unpack_depth)
            else:
                nsource = ops.UnpackArrayOp(sourceedge.source,unpack_depth - pack_depth)
            self.graph.addNode(nsource)
            self.graph.addEdge(query_graph.ParamEdge(sourceedge.source,nsource,"slice"))
        else:
            nsource = sourceedge.source
        
        ttarget_edges = list(self.graph.edge_source[target])
        for edge in ttarget_edges:
            self.graph.dropEdge(edge)
            edge.source = nsource
            self.graph.addEdge(edge)

        self.graph.dropNode(node)
        self.graph.dropNode(target)
           

        



import operator

import manager
import create_graph
from ..utils.multi_visitor import VisitorFactory, NF_ELSE
from ..utils import util
_delay_import_(globals(),"..query_graph")
class SerializeExec(VisitorFactory(prefixes=("visit","assign"), flags=NF_ELSE), manager.Pass):

    after = set([create_graph.CreateGraph])

    @classmethod
    def run(cls, query, run_manager):
        self = cls()
        self.graph = run_manager.pass_results[create_graph.CreateGraph]
        self.commands = []
        self.visit(self.graph.root)
        return self.commands
    
    def visitelse(self,node):
        raise RuntimeError, "Unknown type of node encountered: " + str(type(node))

    def visitNode(self,node):
        params = dict()
        for edge in self.graph.edge_target[node]:
            if(isinstance(edge, query_graph.ParamEdge)):
                source = edge.source
                if(source in self.graph.na['exec_order']):
                    param_idx = self.graph.na['exec_order'][source]
                else:
                    param_idx = self.visit(source)
                self.graph.na['param_usecount'][source] += 1

                self.assign(edge, params, param_idx)
               

        command_id = len(self.commands)
        self.commands.append(node)
        self.graph.na['exec_order'][node] = command_id
        self.graph.na['param_idxs'][node] = params
        self.graph.na['param_usecount'][node] = 0
        return command_id

    def assignParamEdge(self, edge, params, param_idx):
        params[edge.name] = param_idx

    def assignParamListEdge(self, edge, params, param_idx):
        if(not edge.name in params):
            params[edge.name] = [None]
        while(len(params[edge.name]) <= edge.pos):
            params[edge.name].append(None)
        params[edge.name][edge.pos] = param_idx





from ibidas.utils import util
from ibidas.utils.config import config
import os.path
import os
from logging import warning
_delay_import_(globals(),"ibidas","*")

class Pre(object):
    def register(self, func, name=None, category=None):
        if category:
            if not category in self.__dict__:
                self.__dict__[category] = Pre()
            self.__dict__[category].register(func, name)
        else:
            if(name is None):
                name = func.__name__
            setattr(self,name,func)
predefined_sources = Pre()

############################# YEAST ######################################
def yeast_feats():
    """Returns table of yeast genome features,
       from:  http://downloads.yeastgenome.org/curation/chromosomal_feature/SGD_features.tab
    """

    rtype = """[feats:*]<(sgdid=bytes, feat_type=bytes, feat_qual=bytes, feat_name=bytes, gene_name=bytes, 
                          gene_aliases=bytes, feat_parent_name=bytes, sgdid_alias=bytes, chromosome=bytes, 
                          start=bytes, stop=bytes, strand=bytes[1], genetic_pos=bytes, coordinate_version=bytes[10], 
                          sequence_version=bytes, description=bytes)"""
    
    res = Read(Fetch("http://downloads.yeastgenome.org/curation/chromosomal_feature/SGD_features.tab"),dtype=rtype)

    splitfunc = lambda x: x.split("|")
    res = res.To(_.gene_aliases,  Do=_.Each(splitfunc, dtype="[aliases:~]<bytes").Elem()[_ != ""])
    res = res.To(_.start, _.stop, Do=_.Cast("int?"))
    res = res.To(_.genetic_pos,   Do=_.Cast("real64?"))
    return res.Copy()
predefined_sources.register(yeast_feats,name="genomic_feats",category="yeast")  


def yeast_aliases(feats, remove_multi=True):
    """Given a yeast feats table (see ``yeast_feats``), returns a 
       two column table with aliases and the corresponding  official feature names.
       All names in output are in uppercase.

       :param remove_multi: if True (default), removes aliases that map to multiple offical feature names.
       """

    res = Stack(feats.Get(_.gene_aliases, _.feat_name).Flat(), 
                feats.Get(_.gene_name, _.feat_name), 
                feats.Get(_.feat_name, _.feat_name))
    res = res/("alias", "feat_name")
    res = res%"feat_aliases"
    res = res[_.alias != ""].Unique()
    res = res.Each(str.upper, dtype="bytes")
    if remove_multi:
        multi_aliases = res.GroupBy(_.alias)[_.feat_name.Count() > 1].alias
        res = res[~(_.alias |In| multi_aliases)]
    
    return res.Copy()
predefined_sources.register(yeast_aliases,name="name_aliases",category="yeast")  


def yeast_phosphogrid():
    url = "http://www.phosphogrid.org/downloads/phosphosites.txt"
    rtype = """[phosphosites:*]<(orf_name=bytes, gene_name=bytes, phospho_site=bytes, site_evidence=bytes, 
               site_evidence_pubmed=bytes, site_conditions=bytes, site_conditions_pubmed=bytes, site_functions=bytes, site_functions_pubmed=bytes, 
               kinases_orfs=bytes, kinases_genes=bytes, kinases_evidence=bytes, kinases_evidence_pubmed=bytes, phosphatases_orfs=bytes, 
               phosphatases_genes=bytes, phosphatases_evidence=bytes, phosphatases_evidence_pubmed=bytes, sequence=bytes, notes=bytes)"""
    res = Read(Fetch(url),fieldnames=True,dtype=rtype)
    splitfunc = lambda x: x.split("|")
    res = res.To(_.site_evidence,  Do=_.Each(splitfunc, dtype="[site_evidences:~]<bytes").Elem()[_ != "-"])
    res = res.To(_.site_evidence_pubmed,  Do=_.Each(splitfunc, dtype="[site_evidence_pubmeds:~]<bytes").Elem()[_ != "-"])
    res = res.To(_.kinases_orfs,  Do=_.Each(splitfunc, dtype="[kinases:~]<bytes").Elem()[_ != "-"])
    res = res.To(_.phosphatases_orfs,  Do=_.Each(splitfunc, dtype="[phosphatases:~]<bytes").Elem()[_ != "-"])
    return res.Copy()
predefined_sources.register(yeast_phosphogrid,name="phosphogrid",category="yeast")  


def yeast_kinome():
    url = "http://www.yeastkinome.org/files/supplementary/Website_Supplementary_Table_1_-_SAINT_Datasets.zip"
    res = Unpack(Fetch(url),subfiles='RAW/RAW-UNIQUE.txt')

    rtype = "[yeastkinome:*]<(bait_yorf=bytes, hit_yorf=bytes, bait_gene=bytes, hit_gene=bytes, saint=bytes)"
    res = Read(res,dtype=rtype,fieldnames=True)    
    res = res.To(_.saint, Do=_.Cast("real64"))
    return res
predefined_sources.register(yeast_kinome,name="yeast_kinome",category="yeast")  


def yeast_chipchip(translator=None):
    url = "http://fraenkel.mit.edu/improved_map/orfs_by_factor.tar.gz"
    res = Unpack(Fetch(url))
    dirname = os.path.dirname(res[0])
    pval_names = ['none', 'p0.005', 'p0.001']
    max_pvals = [1.0, 0.005, 0.001]
    cons = [0, 1, 2]

    loaded_data = []
    for mpval, pval_name in zip(max_pvals, pval_names):
        for con in cons:
            filename = os.path.join(dirname,"orfs_by_factor_%s_cons%s.txt" % (pval_name, str(con)))
            f = open(filename)
            results = []
            for row in f:
                res = row.split('\t')
                results.append((res[0], res[1:-1]))
            results = Rep(results).Flat()
            results = results.Get(_.f0/"trans_name",_.f1/"target", Rep(mpval)/"pval", Rep(con)/"conservation").Level(1)
            loaded_data.append(results)
    res = Stack(*loaded_data)%"chipchip"

    #group on tf,target tuple, keeping trans_name and target flat
    res = res.GroupBy((_.trans_name, _.target),flat=(_.trans_name, _.target))
    #take min pval found, and max conservation
    res = res.Get(_.trans_name, _.target, _.pval.Min(), _.conservation.Max())   
    if(not translator is None):
        res = res.To(_.trans_name, _.target, Do=_.Each(str.upper,dtype="bytes"))
        res = res.Replace(_.trans_name, translator)
        res = res.Replace(_.target, translator)
    return (res%"chipchip").Copy()
predefined_sources.register(yeast_chipchip,name="chipchip_macisaac",category="yeast")  
        
        
def yeastract(translator=None, url="http://www.yeastract.com/download/RegulationTwoColumnTable_Documented_20111009.tsv.gz"):
    """Downloads documented transcription factor regulation interactions from yeastract"""
    rtype = "[tftargets:*]<(trans_factor=bytes, target=bytes)"
    res = Read(Fetch(url),dtype=rtype)
    if(not translator is None):
        res = res.Each(str.upper,dtype="bytes")
        res = res.Replace(_.trans_factor, translator)
        res = res.Replace(_.target, translator)
    return (res%"yeastract").Copy()
predefined_sources.register(yeastract,category="yeast")

################################ HUMAN ##################################

def omim_genemap():
    """The omim genemap data"""
    rtype = """[omim:*]<(chr_map_entry_nr=bytes, month=bytes, day=bytes, year=bytes, location=bytes, gene_alias=bytes, 
                       gene_status=bytes[1], title=bytes, f8=bytes, mim=bytes, method=bytes, comments=bytes, f12=bytes, 
                       disease1=bytes, disease2=bytes, disease3=bytes, mouse=bytes, reference=bytes)"""

    warning('OMIM FTP is not up to date. See ftp://ftp.ncbi.nih.gov/repository/OMIM/ARCHIVE/README.txt')
    res = Read(Fetch("ftp://ftp.ncbi.nih.gov/repository/OMIM/ARCHIVE/genemap"),dtype=rtype)
    
    res = res.Get(HArray(_.disease1, _.disease2, _.disease3)[_ != " "]/"disease", "~")
    splitfunc = lambda x: x.split(", ")
    res = res.To(_.gene_alias,  Do=_.Each(splitfunc, dtype="[symbols:~]<string").Elem()[_ != ""])
    res = res.To(_.method,      Do=_.Each(splitfunc, dtype="[methods:~]<string").Elem()[_ != ""])
    res = res.To(_.disease,     Do=_.Sum().Each(omim_disease_parse,"[diseases:~]<string").Elem()[_ != ""])
    res = res.To(_.day, _.month, _.year, Do=_.Cast("int?"))
    res = res.To(_.mim, Do=_.Cast("int"))

    return res.Without(_.f8, _.f12, _.disease1, _.disease2, _.disease3).Copy()
predefined_sources.register(omim_genemap, category="human")

def omim_disease_parse(x):
    elems = x.replace('{','').replace('}','').split('; ')
    return [elem.split(', ')[0] for elem in elems]


##################### GENERAL ################################
def in_memory_db():
    """Returns an empty in memory database"""
    return Connect("sqlite:///:memory:");
predefined_sources.register(in_memory_db)

def file_db(filename):
    """Returns an empty in memory database"""
    return Connect("sqlite:///%s" % filename);
predefined_sources.register(file_db)


@util.memoized
def string(dburl=config.get('databases.string_url',None)):
    return Connect(dburl)
predefined_sources.register(string,category="string")

def string_interactions(dburl=config.get('databases.string_url',None), species="Saccharomyces cerevisiae", subscores=False, external_names=False):
    """Given a Postgres db with String data, specified in dburl, and a species, returns all interactions and their score.

    The database data can be obtained from String.

    example url: "postgres://username:password@hostname:port/string_dbname"

    Use ``connect`` to access the whole database::
       
       #Get available species names: 
       >>> connect(dburl).items.species.offical_name

    """
    z = string(dburl)
    inter  = z.items.species |Match| z.network.protein_protein_links
    inter  = inter[_.official_name == species]
    inter  = inter |Match(_.protein_id_a, _.protein_id)| z.items.proteins//"left"
    inter  = inter |Match(_.protein_id_b, _.protein_id)| z.items.proteins//"right"
    if external_names:
        names = inter.Get(_.left.protein_external_id/"left", _.right.protein_external_id/"right").Each(lambda x : x.split('.')[1],dtype="bytes")
    else:
        names = inter.Get(_.left.preferred_name/"left", _.right.preferred_name/"right")

    if(subscores):
        return   inter.Get(names,
                        _.equiv_nscore/"neighborhood_score", _.equiv_nscore_transferred/"neighborhood_score_transferred", 
                        _.equiv_fscore/"fusion_score", 
                        _.equiv_pscore/"phylo_cooccurence_score", 
                        _.equiv_hscore/"homology_score", 
                        _.array_score/"coexpression_score", 
                        _.array_score_transferred/"coexpression_score_transferred",
                        _.experimental_score/"experimental_score",
                        _.experimental_score_transferred/"experimental_score_transferred", 
                        _.database_score/"curated_score", 
                        _.database_score_transferred/"curated_score_transferred", 
                        _.textmining_score/"textmining_score", 
                        _.textmining_score_transferred/"textmining_score_transferred",
                        _.combined_score) % "interactions"
    else:
        return   inter.Get(names, 
                        _.combined_score) % "interactions"
predefined_sources.register(string_interactions,name="interactions", category="string")

def string_interaction_types(dburl=config.get('databases.string_url',None), species="Saccharomyces cerevisiae", external_names=False):
    """Given a Postgres db with String data, specified in dburl, and a species, returns all interactions and their score.

    The database data can be obtained from String.

    example url: "postgres://username:password@hostname:port/string_dbname"

    Use ``connect`` to access the whole database::
       
       #Get available species names: 
       >>> connect(dburl).items.species.offical_name

    """
    z = string(dburl)
    inter  = z.network.actions
    inter  = inter |Match(_.item_id_a, _.protein_id)| z.items.proteins//"left"
    inter  = inter |Match(_.item_id_b, _.protein_id)| z.items.proteins//"right"
    inter = inter |Match(_.left.species_id, _.species_id)| z.items.species
    inter  = inter[_.official_name == species]
    if external_names:
        names = inter.Get(_.left.protein_external_id/"left", _.right.protein_external_id/"right").Each(lambda x : x.split('.')[1],dtype="bytes")
    else:
        names = inter.Get(_.left.preferred_name/"left", _.right.preferred_name/"right")
    return   inter.Get(names,
                    _.mode, _.action, _.a_is_acting,
                    _.score) % "interactions"
predefined_sources.register(string_interaction_types,name="interaction_types",category="string")


@util.memoized
def open_go(dburl=config.get("databases.go_url",None)):
    go = Connect(dburl)
    return go
predefined_sources.register(open_go,name="go",category="go")


def go_annotations(dburl=config.get("databases.go_url",None), genus="Saccharomyces", species="cerevisiae", include_ancestors=False):
    """Accesses GO annotations in a MySQL database.

       Database data can be obtained from the geneontology website.

       example url: "mysql://username:password@hostname:port/go
    """
    go = open_go(dburl)
    g = go.species |Match(_.id,                   _.species_id)|      go.gene_product
    g = g          |Match(_.gene_product.id,      _.gene_product_id)| go.association
    g = g          |Match(_.association.id,       _.association_id)|  go.evidence
    g = g          |Match(_.gene_product.dbxref_id,           _.id)|  go.dbxref
    if include_ancestors:
        g = g          |Match(_.association.term_id, _.term2_id)|     go.graph_path
        g = g          |Match(_.term1_id, _.id)|                      go.term//"annot"
        g = g          |Match(_.relationship_type_id, _.id)|          go.term//"rel"
        g = g[_.rel.name == "is_a"]
    else:
        g = g          |Match(_.association.term_id,   _.id)|         go.term//"annot"
    g = g[(_.genus==genus) & (_.species == species)][_.is_not == False][_.evidence.code != "ND"]
    g = g.ReplaceMissing()
    g = g.Get(_.symbol   /"gene_symbol",   _.xref_key / "gene_id",  
                 _.annot.acc/"go_id",         _.annot.name/"annotation", _.annot.term_type/"go_type", 
                 _.evidence.code/"evidence")%"annotations"
    return g.Copy()
predefined_sources.register(go_annotations,name="annotations",category="go")

def go_info(dburl=config.get("databases.go_url",None), genus="Saccharomyces", species="cerevisiae", include_ancestors=False):
    """Accesses GO term info in a MySQL database.

       Database data can be obtained from the geneontology website.

       example url: "mysql://username:password@hostname:port/go
    """
    go = open_go(dburl)

    #select annotations from a certain species, and their ancestor annotations
    g = go.species
    g = g |Match(_.id,                  _.species_id)|      go.gene_product
    g = g |Match(_.gene_product.id,     _.gene_product_id)| go.association
    g = g |Match(_.association.term_id, _.term2_id)|        go.graph_path
    g = g[(_.genus==genus) & (_.species == species)]

    #link them to terms to get actual go accession id
    g = g |Match(_.term2_id,             _.id)| go.term//"child"
    g = g |Match(_.term1_id,             _.id)| go.term//"parent"
    g = g |Match(_.relationship_type_id, _.id)| go.term//"rel"

    #get relevant fields
    g = g.ReplaceMissing()

    if include_ancestors:
        g = g.Get(_.child.acc/"go_id", _.parent.acc/"parent_id", _.gene_product.id/"gene_id").Copy()
    else:
        g = g.Get(_.child.acc/"go_id", _.parent.acc/"parent_id", _.child.term_type/"go_type", _.child.name/"annotation",
                _.distance, _.rel.name/"relationship", _.gene_product.id/"gene_id").Copy()

    #step A: calculate number of genes associated with go terms
    ngenes = g.GroupBy(_.parent_id).Get(_.parent_id/"go_id",_.gene_id.Unique().Count()/"ngenes")


    if include_ancestors:
        g2 = go.graph_path
        g2 = g2 |Match(_.term2_id,             _.id)| go.term//"child"
        g2 = g2 |Match(_.term1_id,             _.id)| go.term//"parent"
        g2 = g2 |Match(_.relationship_type_id, _.id)| go.term//"rel"
        g2 = g2.Get(_.child.acc/"go_id", _.parent.acc/"parent_id", _.child.term_type/"go_type", _.child.name/"annotation",
                _.distance, _.rel.name/"relationship").Copy()
      
        #step B1: group per term
        g2 = g2[_.go_id |In| ngenes.go_id]
        goinfo = g2.GroupBy(_.go_id, flat=(_.go_type, _.annotation))
        
        #step B2: for ech term, determine type, max depth to root, ancestors (non-unique) and relationship to ancestors
        goinfo = goinfo.Get(_.go_id, _.go_type, _.annotation,  
                            _[_.parent_id == "all"].distance.Max()/"depth",
                            _.parent_id / "ancestor", _.relationship, _.distance)

        #step B3: select for each ancestor the minimum path length
        goinfo = goinfo.GroupBy(_.ancestor)[..., _.distance.Argmin()].Copy()
    else:
        #step B1: group per term
        goinfo = g.GroupBy(_.go_id, flat=(_.go_type, _.annotation))
        
        #step B2: for ech term, determine type, max depth to root, ancestors (non-unique) and relationship to ancestors
        goinfo = goinfo.Get(_.go_id, _.go_type, _.annotation,  
                            _[_.parent_id == "all"].distance.Max()/"depth",
                            _.parent_id / "ancestor", _.relationship, _.distance)

        #step B3: select for each ancestor the minimum path length
        goinfo = goinfo.GroupBy(_.ancestor)[..., _.distance.Argmin()].Copy()

    #step C: combine, copy, return
    return (goinfo |Match| ngenes).Copy()

predefined_sources.register(go_info,name="term_info",category="go")


########################### #KEGG ######################################
@util.memoized
def kegg():
    from SOAPpy import WSDL
    wsdl = 'http://soap.genome.jp/KEGG.wsdl';
    serv = WSDL.Proxy(wsdl)
    return serv

def get_kegg_organisms():
    serv = kegg()
    return Rep(serv.list_organisms())
predefined_sources.register(get_kegg_organisms,name="organisms", category="kegg")

def get_kegg_pathways(org):
    serv = kegg()
    return Rep(serv.list_pathways(org))
predefined_sources.register(get_kegg_pathways,name="pathways", category="kegg")

def get_kegg_pathway(pathway):
    serv = kegg()
    relations = Rep(serv.get_element_relations_by_pathway(pathway))%"relations"
    elements = Rep(serv.get_elements_by_pathway(pathway))%"nodes"
    pwdata = Combine(elements,relations)
    
    names = pwdata.names.Flat().Unique()
    name_info = get_kegg_btit(names)
    res = (pwdata |Match(_.names, _.id, jointype="left")| name_info).Without(_.id)
    #info = res.Get((_.names, _.info)).Each(extract_info,dtype="bytes")/"info"
    #res = res.Without(_.info).Get("*", info)
    return res.Copy()
predefined_sources.register(get_kegg_pathway,name="pathway", category="kegg")

def get_kegg_btit(names):
    serv = kegg()
    names = names()
    name_info = []
    while len(names) > 0:
        xnames = names[:100]
        names = names[100:]
        res = serv.btit(" ".join(xnames))
        for elem in res.split('\n'):
            if(elem):
                pos = elem.index(' ')
                id = elem[:pos]
                info = elem[(pos+1):].split('; ')
                name_info.append((id, info[0],info[1:]))
    dtype = '[idinfo:*]<(id=string, full_name=string, info=[info_elems:~]:string)'
    return Rep(name_info,dtype=dtype)/("id", "full_name","info")
predefined_sources.register(get_kegg_btit,name="id_info", category="kegg")

def extract_info(vals):
    names, info = vals
    res = str(info).split('; ')[0]
    if res == "--":
        res = ""
    return res




"""
The ibidas module contains all main functions for working with ibidas objects.
"""

__all__ = ["Rep","Read","Import","Connect","_","CyNetwork",'Unpack', "Addformat",
           "Array","Tuple","Combine","HArray",
           "Stack","Intersect","Union","Except","Difference",
           "Pos","Argsort","Rank","IsMissing","CumSum",
           "Any","All",
           "Max","Min",
           "Argmin","Argmax",
           "Mean","Median",
           "Sum",
           "Count","Match",
           "Broadcast","CreateType","MatchType",
           "newdim","NEWDIM","LCDIM","LASTCOMMONDIM","COMMON_NAME",'COMMON_POS', "Missing",
           "Corr","In","Contains",
           "Fetch","Serve","Get","Alg",
           "Load","Save",
           "Invert","Abs", "Negative",
           "Add","Subtract","Multiply","Modulo","Divide","FloorDivide","And","Or","Xor","Power","Equal","NotEqual","LessEqual","Less","GreaterEqual","Greater","Each"
           ]

from utils import delay_import
from utils.util import save_rep, load_rep, save_csv
from utils.context import _
from utils.missing import Missing
from utils.infix import Infix
from itypes import createType as CreateType, matchType as MatchType
from wrappers.python import Rep
from constants import *
from repops import Detect
import repops_dim 
from repops_multi import Broadcast, Combine, Sort
import repops_slice 
import repops_funcs
from download_cache import DownloadCache, Unpack
from pre import predefined_sources as Get
from algs import predefined_algs as Alg
from wrappers.cytoscape import CyNetwork
from server import Serve
from constants import *

Fetch = DownloadCache()
In = Infix(repops_funcs.Within)
Contains = Infix(repops_funcs.Contains)
Match = Infix(repops_multi.Match)
Stack = Infix(repops_multi.Stack)
Intersect = Infix(repops_multi.Intersect)
Union = Infix(repops_multi.Union)
Except = Infix(repops_multi.Except)
Difference = Infix(repops_multi.Difference)

Pos = repops.delayable(default_params="#")(repops_funcs.Pos)
IsMissing = repops.delayable()(repops_funcs.IsMissing)
Argsort = repops.delayable()(repops_funcs.Argsort)
Rank = repops.delayable()(repops_funcs.Rank)
CumSum = repops.delayable()(repops_funcs.CumSum)
Argmax = repops.delayable()(repops_funcs.Argmin)
Argmin = repops.delayable()(repops_funcs.Argmax)
Sum = repops.delayable()(repops_funcs.Sum)
Any = repops.delayable()(repops_funcs.Any)
All = repops.delayable()(repops_funcs.All)
Max = repops.delayable()(repops_funcs.Max)
Min = repops.delayable()(repops_funcs.Min)
Mean = repops.delayable()(repops_funcs.Mean)
Median = repops.delayable()(repops_funcs.Median)
Count = repops.delayable()(repops_funcs.Count)
Corr = repops.delayable()(repops_funcs.Corr)
Invert = repops.delayable()(repops_funcs.Invert)
Abs = repops.delayable()(repops_funcs.Abs)
Negative = repops.delayable()(repops_funcs.Negative)


Add = repops.delayable()(repops_funcs.Add)
Subtract = repops.delayable()(repops_funcs.Subtract)
Multiply = repops.delayable()(repops_funcs.Multiply)
Modulo= repops.delayable()(repops_funcs.Modulo)
Divide = repops.delayable()(repops_funcs.Divide)
FloorDivide = repops.delayable()(repops_funcs.FloorDivide)
And = repops.delayable()(repops_funcs.And)
Or = repops.delayable()(repops_funcs.Or)
Xor = repops.delayable()(repops_funcs.Xor)
Power = repops.delayable()(repops_funcs.Power)
Equal = repops.delayable()(repops_funcs.Equal)
NotEqual = repops.delayable()(repops_funcs.NotEqual)
LessEqual = repops.delayable()(repops_funcs.LessEqual)
Less = repops.delayable()(repops_funcs.Less)
GreaterEqual = repops.delayable()(repops_funcs.GreaterEqual)
Greater = repops.delayable()(repops_funcs.Greater)
Each = repops.delayable()(repops_funcs.Each)


HArray = repops.delayable(nsources=UNDEFINED)(repops_slice.HArray)
Tuple = repops.delayable(nsources=UNDEFINED)(repops_slice.Tuple)
Array = repops.delayable()(repops_dim.Array)



def fimport_tsv(url, **kwargs):
    from wrappers.tsv import TSVRepresentor
    return TSVRepresentor(url, **kwargs) 

def fimport_matrixtsv(url, **kwargs):
    from wrappers.matrix_tsv import MatrixTSVRepresentor
    return MatrixTSVRepresentor(url, **kwargs)

def fimport_xml(url, **kwargs):
    from wrappers.xml_wrapper import XMLRepresentor
    return XMLRepresentor(url, **kwargs) 

def fimport_psimi(url, **kwargs):
    from wrappers.psimi import read_psimi
    return read_psimi(url, **kwargs)

def fimport_fasta(url, **kwargs):
    from wrappers.fasta import read_fasta;
    return read_fasta(url, **kwargs);


formats = { 'tsv' : fimport_tsv, 'csv' : fimport_tsv,
            'tsv_matrix' : fimport_matrixtsv,
	    'xml' : fimport_xml,
	    'psimi' : fimport_psimi,
	    'fasta' : fimport_fasta, 'fa' : fimport_fasta, 'fas' : fimport_fasta
	  };

def Addformat(ext, read_fn):
    formats[ext] = read_fn;

def Import(url, **kwargs):

  from os.path import splitext;
  detect=kwargs.pop('detect', True);

  base = url;

  while True:
    (base, ext) = splitext(base);
    ext = ext.split('.')[1] if ext else 'tsv';
    format = kwargs.pop('format', ext).lower();

    if not format:
        raise RuntimeError("Unknown format specified")
    if format not in formats:
        continue;
    else:
        data = formats[format](url, **kwargs);
        return data.Detect() if detect else data;

Read = Import

def Connect(url, **kwargs):
    format = kwargs.pop('format','db')
    if(format == "db"):
        from wrappers.sql import open_db
        return open_db(url, **kwargs)
    else:
        raise RuntimeError("Unknown format specified")


def Save(r, filename):
    if filename.endswith('tsv') or filename.endswith('csv') or filename.endswith('tab'):
        save_csv(r, filename);
    else:
        save_rep(r, filename);


def Load(filename,**kwargs):
    if filename.endswith('tsv') or filename.endswith('csv') or filename.endswith('tab'):
        from wrappers.tsv import TSVRepresentor
        return TSVRepresentor(filename, **kwargs) 
    else:
        return load_rep(filename)

delay_import.perform_delayed_imports()








import copy

class CycleError(RuntimeError):
    pass

class StableTopoSortGraph(object):
    def __init__(self):
        self.nodes = []
        self.parents = {}
        self.ancestors = {}

    def hasNode(self, node):
        return node in self.parents

    def getNodes(self):
        return self.nodes

    def keepNodes(self, nodes):
        nodes = set(nodes)
        
        self.nodes = [n for n in self.nodes if n in nodes]
        self.parents = dict([(k,v & nodes) for k,v in self.parents.iteritems() if k in nodes])
        self.parents = dict([(k,v & nodes) for k,v in self.ancestors.iteritems() if k in nodes])



    def addNodeIfNotExist(self,node):
        if(node in self.parents):
            return False
        else:
            self.nodes.append(node)
            self.parents[node] = set()
            self.ancestors[node] = set()
            return True

    def addNodesIfNotExist(self, *nodes):
        for node in nodes:
            self.addNodeIfNotExist(node)

    def mergeNodes(self,nodeid1,nodeid2):
        if(nodeid1 in self.ancestors[nodeid2] or nodeid2 in self.ancestors[nodeid1]):
            raise CycleError, "Merger introduces cycle!"

        del self.nodes[self.nodes.index(nodeid1)]
        self.parents[nodeid2].update(self.parents[nodeid1])
        self.ancestors[nodeid2].update(self.ancestors[nodeid1])
        del self.parents[nodeid1]
        del self.ancestors[nodeid1]
        for node in self.nodes:
            if nodeid1 in self.parents[node]:
                self.parents[node].discard(nodeid1)
                self.parents[node].add(nodeid2)
            if nodeid1 in self.ancestors[node]:
                self.ancestors[node].discard(nodeid1)
                self.ancestors[node].add(nodeid2)

    def addEdge(self,before,after):
        assert before in self.parents, "Source node not known"
        assert after in self.parents, "Target node not known"

        if after in self.ancestors[before] or after == before:
            raise CycleError, "Edge (%s, %s) introduces cycle!" %(str(before), str(after))

        self.parents[after].add(before)
        self.ancestors[after].add(before)
        self.ancestors[after].update(self.ancestors[before])

    def getParents(self,node):
        return self.parents[node]

    def getAncestors(self,node):
        return self.ancestors[node]

    def getDescendants(self,node):
        return [n for n in nodes if node in self.ancestors[n]]

    def copy(self):
        return copy.deepcopy(self)

    def __iter__(self):
        return StableTopoSortIter(self)

    def fullOrder(self):
        return StableTopoSortFullOrderIter(self)
    
    def elem(self,n,after=None,exclude=None):
        i = self.__iter__()
        if(not after is None):
            for node in i:
                if(node == after):
                    break
            else:
                raise Indexerror, "Not found"

        pos = -1
        for node in i:
            if(exclude is None or not node in exclude):
                pos += 1
            if(pos == n):
                break
        else:
            raise IndexError, "Not found"
        return node


class StableTopoSortIter(object):
    def __init__(self,stable_topo_sort_graph):
        self.sts_graph = stable_topo_sort_graph
        self.visited = set()

    def __iter__(self):
        return self

    def next(self):
        nodes = self.sts_graph.nodes
        visit_node = None
        
        if(len(self.visited) == len(self.sts_graph.nodes)):
            raise StopIteration
        
        for node in nodes:
            if(node in self.visited):
                continue
            parents = self.sts_graph.parents[node]
            if(parents.issubset(self.visited)):
                visit_node = node
                break
        if(visit_node is None):
            raise RuntimeError, "No viable next element during toposort, cycle?"
        self.visited.add(visit_node)
        return visit_node

class StableTopoSortFullOrderIter(StableTopoSortIter):
    def next(self):
        nodes = self.sts_graph.nodes
        visit_nodes = []
        
        if(len(self.visited) == len(self.sts_graph.nodes)):
            raise StopIteration
        
        for node in nodes:
            if(node in self.visited):
                continue
            parents = self.sts_graph.parents[node]
            if(parents.issubset(self.visited)):
                visit_nodes.append(node)
        
        if not visit_nodes:
            raise CycleError, "No viable next element during toposort, cycle?"
                    
        self.visited.update(visit_nodes)
        return visit_nodes

class StableTopoSortCopyIter(StableTopoSortIter):
    def __init__(self,stable_topo_sort_graph):
        self.sts_graph = stable_topo_sort_graph.copy()
        self.visited = set()

    def __iter__(self):
        return self



import copy
import manager
from ..utils import util
from ..utils.multi_visitor import VisitorFactory, NF_ERROR, F_CACHE
from .. import query_graph
from ..constants import *
_delay_import_(globals(),"..repops")
_delay_import_(globals(),"..repops_multi")
_delay_import_(globals(),"..representor")
_delay_import_(globals(),"..ops")

class EnsureInfo(VisitorFactory(prefixes=("findFirstKnown","processQuery"), 
                                      flags=NF_ERROR | F_CACHE), manager.Pass):
    @classmethod
    def run(cls, query, run_manager):
        query_root = query.root
        if(not query_root._slicesKnown()):
            while(not query_root._slicesKnown()):
                self = cls()
                self.copied_nodes = dict()
                first_known_nodes = self.findFirstKnown(query_root)
                temp_root = repops_multi.Combine(*first_known_nodes)
                temp_root = repops.ApplyFuncRep(temp_root,repops.apply_slice,ops.DetectTypeOp,None)
                known_slices = temp_root._getResultSlices(endpoint=False)

                self.inferred_node_slices = dict()
                for node in first_known_nodes:
                    nslices = len(node._slices)
                    self.inferred_node_slices[id(node)] = known_slices[:nslices]
                    known_slices = known_slices[nslices:]
                query_root = self.processQuery(query_root)
            query.root = query_root
        return query.root

    def findFirstKnownRepresentor(self,node):
        if node._slicesKnown():
            return (node,)
        else:
            raise RuntimeError, "Cannot find known slices for " + str(type(node))

    def findFirstKnownUnaryOpRep(self,node):
        if node._slicesKnown():
            return (node,)
        else:
            return self.findFirstKnown(node._source)
    
    def findFirstKnownRequestUnaryOpRep(self,node):
        if node._slicesKnown():
            return (node,)
        else:
            x = list(self.findFirstKnown(node._source))
            if(node._req_sources):
               for name, elem in node._req_sources:
                    if(isinstance(elem, representor.Representor)):
                        x.extend(self.findFirstKnown(elem))
            return tuple(x)
   
    def findFirstKnownMultiOpRep(self,node):
        if node._slicesKnown():
            return (node,)
        else:
            return sum([self.findFirstKnown(source) for source in node._sources],())
    
    def findFirstKnownMultiOpRep(self,node):
        if node._slicesKnown():
            return (node,)
        else:
            return sum([self.findFirstKnown(source) for source in node._sources],())

    def getNodeCopy(self,node):
        if(not id(node) in self.copied_nodes):
            self.copied_nodes[id(node)] = copy.copy(node)
        return self.copied_nodes[id(node)]

    def processQueryRepresentor(self,node):
        if node._slicesKnown():
            assert id(node) in self.inferred_node_slices, "Cannot find node in inferred nodes map"
            nnode = self.getNodeCopy(node)
            nnode._initialize(self.inferred_node_slices[id(node)])
            return nnode
        else:
            raise RuntimeError, "Cannot find known slices for " + str(type(node))

    def processQueryUnaryOpRep(self,node):
        if node._slicesKnown():
            return self.processQueryRepresentor(node)
        else:
            nnode = self.getNodeCopy(node)
            nnode._source = self.processQuery(node._source)
            nnode._process(nnode._source,*node._params[0], **node._params[1])
            return nnode
    
    def processQueryRequestUnaryOpRep(self,node):
        if node._slicesKnown():
            return self.processQueryRepresentor(node)
        else:
            nnode = self.getNodeCopy(node)
            nnode._source = self.processQuery(node._source)
            if(nnode._req_sources):
                npsources = []
                for name,elem in nnode._req_sources:
                     if(isinstance(elem,representor.Representor)):
                        elem = self.processQuery(elem)
                     npsources.append((name,elem))
                nnode._req_sources = npsources
            nnode._process(nnode._source,*node._params[0], **node._params[1])
            return nnode
   
    def processQueryMultiOpRep(self,node):
        if node._slicesKnown():
            return self.processQueryRepresentor(node)
        else:
            nnode = self.getNodeCopy(node)
            nnode._sources = tuple([self.processQuery(source) for source in node._sources])
            nnode._process(nnode._sources,*node._params[0], **node._params[1])
            return nnode

    def processQueryMatch(self, node):
        if node._slicesKnown():
            return self.processQueryRepresentor(node)
        else:
            nnode = self.getNodeCopy(node)
            lsource, rsource, lslice, rslice = node._sources
            if(lslice is None or rslice is None):
                lsource = self.processQuery(lsource)
                rsource = self.processQuery(rsource)
                nnode._process((lsource,rsource,lslice,rslice),*node._params[0], **node._params[1])
                lsource, rsource, lslice, rslice = node._sources

            assert not lslice is None and not rslice is None, "Cannot determine left and/or right slice for Match operation"
            nnode._sources = tuple([self.processQuery(source) for source in node._sources])
            nnode._process(nnode._sources,*node._params[0], **node._params[1])
            return nnode




import sys

req_version = (2,6)
cur_version = sys.version_info

if not ((cur_version[0] >= req_version[0]) or (cur_version[0] == req_version[0] and cur_version[1] >= req_version[1])):
    sys.stderr.write("Your python interpreter is too old. Ibidas needs at least Python 2.6. Please consider upgrading.\n")
    sys.exit(-1)

if cur_version[0] > 2:
    sys.stderr.write("Ibidas only works with Python 2.x. Use 'python2' instead of 'python' to run the setup script.\n")
    sys.exit(-1)
    
import ez_setup
ez_setup.use_setuptools()

from setuptools import setup,find_packages,Extension
import distutils.sysconfig
import os
import os.path

include_dir = distutils.sysconfig.get_python_lib() + "/numpy/core/include/"
if(not os.path.isfile(os.path.join(include_dir, "numpy/arrayobject.h"))):
    #print os.path.join(include_dir, "numpy/arrayobject.h")
    include_dir = include_dir.replace('lib', 'lib64')
    if(not os.path.isfile(os.path.join(include_dir, "numpy/arrayobject.h"))):
        #print os.path.join(include_dir, "numpy/arrayobject.h")
        raise RuntimeError('numpy array headers not found')

if not os.path.isdir('docs/_build'):
    os.mkdir('docs/_build')

setup(
    name="Ibidas",
    version="0.1.21",
    packages = find_packages(),
    test_suite = "test",
    scripts = ['bin/ibidas','bin/ibidas_shell.py','bin/python_2_please'],
    ext_modules = [
        Extension(
    		'closure', 
    		['src/closure.c'],
    		extra_compile_args=["-Wall"]
	    ),
        Extension(
		    'ibidas.utils.multi_visitor', 
		    ['src/multi_visitor.c'],
		    extra_compile_args=["-Wall"]
	    ),
        Extension(
            'ibidas.utils.cutils', 
            ['src/cutils.c'], 
            include_dirs=[include_dir],
            extra_compile_args=["-Wall"]
        ),
     ],
     install_requires=['numpy>=1.4.1','numpy!=1.6.0','sqlalchemy>=0.6.4','ipython>=0.10.1','sphinx>=1.0.5'],
     author = "M. Hulsman & J.J.Bot",
     author_email = "m.hulsman@tudelft.nl",
     description = "Ibidas is an environment for data handling and exploration, able to cope with different data structures and sources",
     url = "https://trac.nbic.nl/ibidas/",
     license = "LGPLv2.1",

)




import os
import csv
from logging import info, warning

import wrapper
from ..itypes import rtypes
from .. import ops
from ..constants import *

import python
from ..itypes import detector
from ..utils import nested_array, util

class TSVRepresentor(wrapper.SourceRepresentor):
    def __init__(self, filename, dialect=False, skiprows=-1, dtype=rtypes.unknown, fieldnames=None, delimiter=None):
        file = util.open_file(filename,mode='rU')
  
        reader, dialect = self.getReader(file, dialect, delimiter=delimiter)
        
        #determine number of rows to skip (e.g. comments)
        if(skiprows == -1):
            skiprows = self.determineSkipRows(file, reader)

        for i in xrange(skiprows):
            file.readline()
        startpos = file.tell()

        #determine type
        if(dtype is None):
            dtype = rtypes.unknown
        elif(dtype == rtypes.unknown):
            if(fieldnames is None):
                sample = []
                for line in file:
                    sample.append(line)
                    if(len(sample) > 10):
                        break
                file.seek(startpos)
                fieldnames = file.readline()
                fieldnames = csv.reader([fieldnames],dialect=dialect).next()
                if(not csv.Sniffer().has_header("\n".join(sample))):
                    file.seek(startpos)
                    fieldnames = None
            elif(fieldnames is True):
                fieldnames = file.readline()
                fieldnames = csv.reader([fieldnames],dialect=dialect).next()
            
            if(fieldnames):
                fieldnames = [util.valid_name(fieldname) for fieldname in fieldnames]

            startpos = file.tell()
            #parse data
            data = [tuple(row) for row in reader]
            file.seek(startpos)

            det = detector.Detector()
            det.process(data)
            dtype = det.getType()
            if(not fieldnames is None and not fieldnames is False and dtype != rtypes.unknown):
                assert isinstance(dtype,rtypes.TypeArray),"Error while determining type"
                assert isinstance(dtype.subtypes[0],rtypes.TypeTuple),"Error while determining type"
                dtype.subtypes[0].fieldnames = tuple(fieldnames)
        elif(isinstance(dtype,basestring)):
            dtype = rtypes.createType(dtype)
            if(fieldnames is True):
                dummy = file.readline()
                startpos = file.tell()
        else:
            raise RuntimeError, "Not a valid type specified"

        slice = TSVOp(filename, dialect, startpos, dtype, "data")
        if(slice.type.__class__ is rtypes.TypeArray):
            slice = ops.UnpackArrayOp(slice)
        if(slice.type.__class__ is rtypes.TypeTuple):
            nslices = [ops.UnpackTupleOp(slice, idx) for idx in range(len(slice.type.subtypes))]
        else:
            nslices = [slice]
 
        file.close()
        self._initialize(tuple(nslices))

    def getReader(self, file, dialect=False, delimiter=None):
        #determine dialect, create csv parser
        if(dialect is False):
            lines = []
            #get sample
            for line in file:
                lines.append(line)
                if(len(lines) > 500):
                    break
            #sniff last 20 lines
            lines = lines[-20:]
            dialect = csv.Sniffer().sniff("\n".join(lines),delimiters=delimiter)
            file.seek(0)
            reader = csv.reader(file, dialect)
        else:
            if(issubclass(dialect, csv.Dialect)):
                reader = csv.reader(file, dialect)
            else:
                reader = csv.reader(file, delimiter=dialect)
        return (reader, dialect)

    def determineSkipRows(self, file, reader):
        splitsize = []
        for row in reader:
            splitsize.append(len(row))
            if(not row or len(row[0]) == 0 or ((len(splitsize) == 1 or splitsize[-2] == 0) and (row[0][0] == '#' or row[0][0] == '%' or row[0][0] == "!"))):
                splitsize[-1] = 0
            elif(len(splitsize) > 75):
                x = set(splitsize[-20:])
                if(len(x) == 1 and not x.pop() == 1):
                    break
            elif(len(splitsize) > 500):
                raise RuntimeError, "Cannot find correct number of columns. Incorrect delimiter?" 
        
        real_split  = splitsize[-1]
        skiprows = 0
        for pos, split in enumerate(splitsize):
            if(split != real_split):
                skiprows = pos + 1
        info('Skipping first %d rows', skiprows)                
        file.seek(0)
        return skiprows        

class TSVOp(ops.ExtendOp):
    __slots__ = ["filename", "dialect","startpos"]

    def __init__(self, filename, dialect, startpos, rtype, name):
        self.filename = filename
        self.dialect = dialect
        self.startpos = startpos
        ops.ExtendOp.__init__(self,name=name,rtype=rtype)

    def py_exec(self):
        file = util.open_file(self.filename,mode='rU')
        file.seek(self.startpos)

        if(issubclass(self.dialect, csv.Dialect)):
            reader = csv.reader(file, self.dialect)
        else:
            reader = csv.reader(file, delimiter=self.dialect)
        data = [tuple(row) for row in reader]
        file.close()
        ndata = nested_array.NestedArray(data,self.type)
        return python.ResultOp.from_slice(ndata,self)




import unittest
import numpy

from ibidas.utils import util
from ibidas import *


class TestBugs(unittest.TestCase):
    def test_posbroadcast_match(self):
        yeastract = Get.yeast.yeastract()
        yeastract = yeastract.GroupBy(_.trans_factor)[:10].Copy()
        str(yeastract |Match(_.target)| yeastract[:,newdim])
        str(yeastract |Match(_.target,mode="pos")| yeastract[:,newdim])

    
    def test_filterafter_complexmatch(self):
        yeastract = Get.yeast.yeastract()
        yeastract = yeastract.GroupBy(_.trans_factor)[:10].Copy()
        res = yeastract |Match(_.target,mode="pos")| yeastract[:,newdim]
   
        str(res[0,0,0])
        str(res[:,:,0])
        self.assertTrue(all(All((res[0,...] ==+ res[0,:,:]).FlatAll())()))
        str(res[:2,...])
        str(res[[0,1],...])
        str(res[Rep([0,1]),...])
        str(res[_.target == "YGR258c"])

    def test_posbroadcast(self):
        x = Rep([1,2,3])
        self.assertTrue((x ++ x).Depth == 1)
        self.assertTrue((x ++ x[:,newdim]).Depth == 2)
        self.assertTrue((x ++ x[...,newdim]).Depth == 2)

    def test_filter(self):
        yeastract = Get.yeast.yeastract()
        y = yeastract.GroupBy(_.trans_factor)[:10].Copy()
        str(y.Get(_.trans_factor, _.target[:,newdim][:10]).Count())
    
    def test_slicesunknownmatch(self):
        x = Rep(['test']).Each(str.upper)
        str(x |Match(_.Each(str.upper))| x)


    def test_filter_on_transposed_nested(self):
        z = Rep((['a','b','b','c'],[[1,2,3,4],[1,2,3,4],[4,3,2,1]]))
        res = z.To(_.f1, Do=_.Transpose()).GroupBy(_.f0).To(_.f1, Do=_.Mean(dim=1)).To(_.f1, Do=_.Transpose()).Filter(slice(0,5),dim=1)
        str(res)

    def test_add_on_matrix_plus_nested(self):
        data = [([6,2],[0.5]), ([3,4], [0.3,0.4]), ([6,4], [0.8,0.2])]
        z = Rep(data)
        str(z.f0 + z.f1)
        str(z.f1 + z.f0)
        str(z.f0.Transpose() + z.f1)
        str(z.f1 + z.f0.Transpose())

    def test_add_nothing(self):
        data = [(163818, 159108, 455, [[  9, 455]]), (181601, 159108, 748, [[  2, 559],  [  7, 212],  [ 13, 364]])];
        z    = Rep(data);
        db   = Get.in_memory_db();
        db.Store('t', z[:0]);




import os

import wrapper
_delay_import_(globals(),"wrapper_py","Result","rep")
_delay_import_(globals(),"slices")
_delay_import_(globals(),"..itypes","rtypes")


def read_chipchip(path):
    files = os.listdir(path)
    pval_names = ['none', 'p0.005', 'p0.001']
    max_pvals = [1.0, 0.005, 0.001]
    cons = [0, 1, 2]

    loaded_files = []
    for mpval, pval_name in zip(max_pvals, pval_names):
        for con in cons:
            filename = path + "orfs_by_factor_%s_cons%s.txt" % (pval_name, str(con))
            res = ChipChipRepresentor(filename)
            res = res.E.A.join(Rep([(mpval, con)])/("max_pval", "conservation"))
            res = res.flat().tuple().array()
            loaded_files.append(res)

    res = sum(loaded_files[1:], loaded_files[0]).E.A
    res = res.group_by(_.regulator, name="orfs")
    res = res.get(_.regulator, _.Dorfs.group_by(_.orfs))
    res = res.get(_.regulator, _.orfs, _.max_pval.min(), _.conservation.max())

    return res.copy()
    

cctype = "[regulators:*](tuple(regulator=string[], orfs=[orfs:~](string[])))"

class ChipChipRepresentor(wrapper.SourceRepresentor):
    _select_indicator = None
    _select_executor = None
    
    def __init__(self, filename):
        self._filename = filename 
        
        ntype = rtypes.createType(cctype)
        tslices = (slices.Slice("data", ntype),)

        all_slices = dict([(slice.id, slice) for slice in tslices])
        wrapper.SourceRepresentor.__init__(self, all_slices, tslices)

    def pyexec(self, executor):
        results = []
        file = open(self._filename)
        for row in file:
            res = row.split('\t')
            results.append((res[0], res[1:-1]))

        res = {}
        res[self._active_slices[0].id] = results
        return Result(res)





import sys
import copy
import itertools
from collections import defaultdict
import numpy

from query_graph import Node
from itypes import rtypes
from constants import *
from thirdparty import tableprint, console

_delay_import_(globals(),"utils","util","context","infix")
_delay_import_(globals(),"itypes", "dimensions","dimpaths")
_delay_import_(globals(),"wrappers","python","wrapper")
_delay_import_(globals(),"query_context")
_delay_import_(globals(),"engines")
_delay_import_(globals(),"repops")
_delay_import_(globals(),"repops_multi")
_delay_import_(globals(),"repops_dim")
_delay_import_(globals(),"repops_slice")
_delay_import_(globals(),"repops_funcs")
_delay_import_(globals(),"ops")




class Representor(Node):
    """Representor is the primary object in Ibidas. It represents
    a data set, accesible through slices.

    Properties can be accessed to obtain information about this object:
    
        * Names: Slice names

        * Type: Data type

        * Slices: List of slices

        * I: Info on slices/types/dims without executing the query

        * Depth: Maximum number of dimensions in slices

    Slices can be accessed as attributes, e.g: obj.slicename

    Note that all slice names should follow the python syntax rules for variable
    names, AND use only lower case letters (to distinguish them from method names, 
    which start all with an uppercase letter). 

    Special attribute access can be obtained through so-called axis specifiers:

        * Bookmarks:  obj.Bbookmarkname   Access set of slices with certain bookmark (see Bookmark method)
        
        * Dimensions: obj.Ddimname        Access all slices with a certain dimension

        * Elements:   obj.E[dimname]      Access Elements of packed arrays. Optional dimname specifies which dimensions to unpack.
                                          Slices without that dimension as outermost dimension are not unpacked. 

        * Fields:     obj.Ffieldname      Access Fields of packed tuples. obj should have only one slice. 

        * Left/Right: obj.L, obj.R        Special nested bookmarks set by e.g. Match operation to allow 
                                          backtracking to separate sources. eg::
                                          
                                          >>> ((x |Match| y) |Match| z).LR  
                                          
                                          gives all slices of y (first go Left (get xy), then R (get y)). 

    """

    _state = 0                 #default value
    _slices = []               #default value

    def _initialize(self, slices):
        assert isinstance(slices, tuple), "slices should be a tuple"
        self._slices = slices

    def _typesKnown(self):
        return isinstance(self._slices, tuple) and not any([slice.type == rtypes.unknown for slice in self._slices])

    def _slicesKnown(self):
        return isinstance(self._slices,tuple)

    def _checkState(self,check_type=False):
        if(check_type is True):
            found = self._typesKnown()
        else:
            found = self._slicesKnown()

        if(not found):
            self._initialize(self._getResultSlices(endpoint=False))

    def Show(self,table_length=100):
        """Prints table of contents.

        :param table_length: Number of rows to show for each dimension (default: 100)

        Show can be used to view a larger part of the table then the default output (15 rows)
        you get in Ipython/Ibidas or by using str(). 

        Show returns its representor object, allowing you to include it at any point in a query
        to observe results, e.g.::

        >>> x.Unique().Show() |Match| y 

        """
        nself = repops.NoOp(self)
        nself._table_length = table_length
        return nself

    def __str__(self, print_data=True, table_length=None):
        self._checkState()

        if table_length is None:
            if '_table_length' in self.__dict__:
                table_length = self._table_length
            else:
                table_length = 15
        
        names = ["Slices:"] + [s.name for s in self._slices]
        types = ["Type:"] + [str(s.type) for s in self._slices]
        longdims = ["Dims:"]
        for s in self._slices:
            dim_str = []
            for  dim in s.dims:
                dim_str.append(str(dim))
            dim_str = "<".join(dim_str)
            longdims.append(dim_str)
                
        dims = ["Dims:"]
        last_dim = ()
        for s in self._slices:
            dim_str = []
            for pos, dim in enumerate(s.dims):
                if len(last_dim) > pos and dim == last_dim[pos]:
                    dim_str.append(".")
                else:
                    dim_str.append(str(dim))
            last_dim = s.dims
            dim_str = "<".join(dim_str)
            dims.append(dim_str)
        
        rows = [names, types, longdims]
       
        opath = dimpaths.getOrderDim([s.dims for s in self._slices])


        if(print_data):
            first_dims = set([s.dims[0] for s in self._slices if s.dims])
            if(first_dims):
                for dim in first_dims:
                    self = self.Filter(slice(None,table_length + 1),dim=dim)
                data = self()
            else:
                data = self()
            if(len(self._slices) == 1):
                data = (data,)

            cols = [[""] * (table_length + 1)]
            maxlen = 0
            
            maxwidth = [0]
            for dcol,s in zip(data, self._slices):
                if(len(s.dims) >= 1):
                    col = [str(row).replace('\n','; ') for row in dcol[:table_length]]
                    maxlen = max(len(col),maxlen)
                    if(len(dcol) > table_length):
                        col[table_length-1] = "..."
                    elif(len(col) < table_length):
                        col.extend([""] * (table_length - len(col)))
                        
                else:
                    col = [""] * (table_length + 1)
                    col[0] = str(dcol)
                    maxlen = max(1,maxlen)
                cols.append(col)
            
            rows.append([""] * len(rows[0]))
            rows[-1][0] = "Data:"
            for i in range(maxlen):
                row = [col[i] for col in cols]
                rows.append(row)
        
        console_height, console_width = console.getTerminalSize()
        widths = tableprint.calc_width(rows)
        indices = tableprint.advise_splits(console_width, widths, 3)

        res = ""
        for i in range(len(indices)):
            srows = tableprint.select_cols(rows, indices[i])
            if(i > 0):
                res += "\n\n"
                col0 = tableprint.select_cols(rows,0)
                srows = tableprint.prepend_col(srows,col0)
            widths = tableprint.calc_width(srows)
            nwidths = tableprint.optimize_width(console_width, widths, 3)
            srows = tableprint.row_crop(nwidths, srows)
            res += tableprint.indent(srows,hasHeader=True) 
        
        if opath:
            dim_str = []
            for  dim in opath:
                dim_str.append(str(dim))
            dim_str = "<".join(dim_str)
            res = res + '\nDim order: ' + dim_str 

        return res 

    def _getInfo(self):
        class Info(object):
            def __repr__(self):
                return self.info
        k = Info()
        k.info = self.__str__(False)
        return k
    I=property(fget=_getInfo)
    Info=I

    #def __repr__(self):
    #    return str(self.__class__)
    __repr__ = __str__

    def _axisF(self, name): 
        if(name == ""):
            name = None
        return repops_slice.UnpackTuple(self, name)

    def _axisE(self, name):
        if(name == ""):
            name = None
        return repops_dim.UnpackArray(self, name)

    def _axisD(self, name):
        #slices = self._active_dim_slice_dict[name] 
        return repops_slice.ProjectDim(self, name)

    def _axisB(self, name):
        return repops_slice.ProjectBookmark(self,name)

    def _axisL(self, name):
        bname = "!L"
        while(name and (name[0] == "L" or name[0] == "R")):
            bname = bname + name[0]
            name = name[1:]

        r = repops_slice.ProjectBookmark(self,bname)
        if(name):
            r = self.__getattr__(name)
        return r
        
    def _axisR(self, name):
        bname = "!R"
        while(name and (name[0] == "L" or name[0] == "R")):
            bname = bname + name[0]
            name = name[1:]

        r = repops_slice.ProjectBookmark(self,bname)
        if(name):
            r = self.__getattr__(name)
        return r

    def __getattr__(self, name):
        if(not name):
            return self

        axis_letter = name[0]
        #should be upper case
        if(not axis_letter.upper() == axis_letter or axis_letter == "_"):
            try:
                return repops_slice.Project(self,name)
            except AttributeError, error:
                #reraise attribute error as runtime error, 
                #so that python will not
                #attempt to find attribute in another way
                #(thereby raising a new exception)
                exc_class, exc, traceback = sys.exc_info()
                raise RuntimeError, RuntimeError(error.message), traceback
        elif(hasattr(self, '_axis' + axis_letter)):
            try:
                name = name[1:]
                return getattr(self, "_axis" + axis_letter)(name)
            except AttributeError, error:
                #reraise attribute error as runtime error, 
                #so that python will not
                #attempt to find attribute in another way
                #(thereby raising a new exception)
                exc_class, exc, traceback = sys.exc_info()
                raise RuntimeError, RuntimeError(error.message + " (and there is also no method '" + axis_letter + name + "')"), traceback
        else:
            raise RuntimeError, 'No method with name: ' + name

    def _getAttributeNames(self):
        if not self._slicesKnown():
            return []
        else:
            #print ""
            #print self.__str__(False)
            return [slice.name for slice in self._slices]

    def Copy(self, log=False, debug=False):
        """Executes the current query.

        Normally, query operations (e.g. obj + 3) are not executed immediately. Instead
        these operations are performed simultaneously when output is requested. This allows
        us to optimize these operations all together, or e.g. translate them into a SQL
        query. 

        However, this behaviour is not always what is needed, e.g::

            >>> x = very expensive query
            >>> print x[10:20]
            >>> print x[10:30]
            
        would execute the query saved in x two times (as ibidas due to being part of an 
        interpreted language cannot analyze the whole script to determine that the output is 
        required twice). 

        To prevent this, one can instead do::

            >>> x = (very expensive query).Copy()

        executing the expensive part of the only query once. 


        :param log:  Setting this to true will print the amount of time that is spent in any of the passes of
                     they query optimizer (default: False)

        :param debug: Setting this to true will output the query tree at various stadio through XML-RPC for visualization
                      in Cytoscape. This requires that Cytoscape is running, with an activated XML-RPC plugin listening
                      at port 9000. 
                     

        """

        res = wrapper.SourceRepresentor()
        res._initialize(tuple(self._getResultSlices(endpoint=False, log=log, debug=debug)))
        return res

   
    def __reduce__(self):
        self = self.Copy()
        slices = self._getResultSlices(endpoint=False)
        for slice in slices:
            slice.source = None
        return (python.PyRepresentor, (slices,))

    def __copy__(self):
        nself = Representor()
        nself.__class__ = self.__class__
        nself.__dict__ = self.__dict__.copy()
        return nself

   
    def _getType(self):
        """Returns type of this object.
           If multiple slices, returns tuple type of slice types.
        """
        if(len(self._slices) == 1):
            return self._slices[0].type
        else:
            return rtypes.TypeTuple(False, 
                    tuple([slice.type for slice in self._slices]), 
                    tuple([slice.name for slice in self._slices]))
    Type=property(fget=_getType)

    def _getSlices(self):
        self._checkState()
        return list(self._slices)
    Slices=property(fget=_getSlices)

    def _getDepth(self):
        """Returns max dimension depth (number of dimensins) of
           slices in this representor. 
        """
        return max([len(slice.dims) for slice in self._slices])
    Depth=property(fget=_getDepth)
    

    def _getNames(self):
        self._checkState()
        return [slice.name for slice in self._slices]
    Names=property(fget=_getNames)

    def __getitem__(self, condition):
        self._checkState()
        if(not isinstance(condition, tuple)):
            condition = (condition,)
        
        #add dimensions first
        ipos = 0
        for pos, cond in enumerate(condition):
            if(isinstance(cond,NewDim)):
                self = repops_dim.InsertDim(self,pos + ipos, cond.name)
            elif(cond is Ellipsis):
                ulength = len(dimpaths.uniqueDimPath([s.dims for s in self._slices], only_unique=True))
                rem_length = len([c for c in condition[(pos + 1):] if not isinstance(cond,NewDim)])
                newnextpos = ulength - rem_length
                curnextpos = pos + ipos
                ipos += newnextpos - curnextpos #skip some dims
        
        #next, perform filtering in backwards order
        #note: cannot use Ellipsis in condition, will perform Ellipsis == elem,
        #which wille xecute as query and fail/be slow
        if(any([elem is Ellipsis for elem in condition])):  
            ncond = len(dimpaths.getOrderDim([s.dims for s in self._slices]))
        else:
            ncond = len(condition)
        
        for pos, cond in enumerate(condition[::-1]):
            if(isinstance(cond, context.Context)):
                cond = context._apply(cond, self)
            if(isinstance(cond,slice) and cond.start is None and 
                                                cond.stop is None and 
                                                cond.step is None):
                pass
            elif(cond is Ellipsis):
                ncond = len(condition)
            elif(isinstance(cond,NewDim)):
                pass
            elif(len(condition) == 1):
                self = repops_multi.Filter(self, cond)
            else:
                self = repops_multi.Filter(self, cond, ncond - pos - 1)
        return self
     
    def Redim(self, *args, **kwds):
        """Assign new dimensions

        example: .Redim('new_dimname', _.f0)
        Assign new dim with name 'new_dimname' to first
        dimension of slice f0

        example: .Redim('new_dimname', f0=1)
        Assign new dim with name 'new_dimname' to second
        dimension of slice f0

        example: .Redim('new_dimname', {_.Dd1.Without('f0'):1, 'f0':1})
        Assign all slices with dimension d1 (except f0) as first dim
        a new dim with name 'new_dimname'. Do the same to slice f0, 
        but as second dimension. 
        """
        return repops_dim.Redim(self, *args, **kwds)
          
    def Filter(self, condition, dim=LASTCOMMONDIM, mode=None):
        """Performs filtering on this dataset using ``condition``.
           
           :param condition: condition to filter on

                * Non-representor values are converted using ``Rep`` function

                * condition should have only a single slice.

                Various data types can be used:

                * Bool: last dim should be equal to a dim in this representor. Is applied to that dim by default.

                * Integer: collapses the dimension it is applied on. 

                * Array (of integers): selects positions indicated by integers in array.

                * Slice: selects slice from array.

           :param dim: Dim to apply the filtering on. 

                * If no dim given, applied to last common dimension of slices (except for bool types).

                * Integer: identifies dimension according to dim order (printed at the end of a representor printout)

                * Long: identifies dimension according to common dimensions shared by all slices (default: -1)

                * String: dimension name

                * Dim object: x.Slices[0].dims[2]

                * Dimpath object: e.g. x.Slices[0].dims[:2]

            :param mode: Determines broadcasting method. 

                * "pos"  Postion-based broadcasting ('numpy'-like broadcasting)

                * "dim"  Dimension-identity based broadcasting (normal 'ibidas' broadcasting)

                * None:  (Default). Determined based on input. Non-representation objects use position-based broadcasting
                         Representation objects by default use dimension-based, except if they are prepended by a '+' operator, 
                         e.g::

                         >>> x.Filter(+constraintrep)

            What is done to dimensions in the constraint that are not in the data source? Here we follow
            the default rules in Ibidas for broadcasting. 

                * First, the dimension in the source that is going to be filtered is identified (according to dim param or constraint last dimension)
                
                * Secondly, we match this dimension to the last dimension in the constraint. 
                
                * All remaining dimensions are broadcasted against each other.

            Examples::

                >>> x = Rep([[1,2,3],[4,5,6]]) 
                Slices: | data     
                -------------------
                Type:   | int64    
                Dims:   | d1:2<d2:3
                Data:   |          
                        | [1 2 3]  
                        | [4 5 6]  
                
                Dim order: d1:2<d2:3
                
                >>> x.Filter(0) 
                Slices: | data
                ---------------
                Type:   | int64
                Dims:   | d1:2
                Data:   |
                        | 1
                        | 4
                
                Dim order: d1:2

            This example matches the last common dimension (d2), and selects the first element.
            This collapses dimension d2. Instead we can also filter on dimension d1::

                >>> x.Filter(0, dim='d1')
                Slices: | data 
                ---------------
                Type:   | int64
                Dims:   | d2:3 
                Data:   |      
                        | 1    
                        | 2    
                        | 3    

                Dim order: d2:3
        """
        if(isinstance(condition, context.Context)):
            condition = context._apply(condition, self)
        return repops_multi.Filter(self, condition, dim, mode) 

    def _getResultSlices(self, args={}, endpoint=True, log=False, debug=False):
        query = query_context.QueryContext(self, args, endpoint)
        if debug:
            return tuple(engines.debug_engine.run(query, log))
        else:
            return tuple(engines.select_engine.run(query, log))

    def __call__(self, **args):
        res = self._getResultSlices(**args)

        if(len(res) == 1):
            return res[0].data
        else:
            return tuple([slice.data for slice in res])

    def Detect(self, *args, **kwargs):
        """Detects types fo slices, and casts result to this type
            
            :param only_unknown: Only detect slices with unknown types [default: False]

            :param allow_convert: Convert also types (e.g. bytes to integers/floats where possible) [default: True]

        """
        return repops.Detect(self, *args, **kwargs).Copy()

    # Overloaded functions. The r* functions are implemented because you
    # want to the same behaviour no matter no which side the known object
    # resides. (a + b == b + a)
    # add ( + )
    def __add__(self, other):
        if(isinstance(other, context.Context)):
            return other.__radd__(self)
        return repops_funcs.Add(self, other)
    
    def __radd__(self, other):
        return repops_funcs.Add(other, self)
    
    def __sub__(self, other):
        if(isinstance(other, context.Context)):
            return other.__radd__(self)
        return repops_funcs.Subtract(self, other)
    
    def __rsub__(self, other):
        return repops_funcs.Subtract(other, self)

    # multiplication ( * )
    def __mul__(self, other):
        if(isinstance(other, context.Context)):
            return other.__rmul__(self)
        return repops_funcs.Multiply(self, other)
    
    def __rmul__(self, other):
        return repops_funcs.Multiply(other, self)
    
    # modulo ( % )
    def __mod__(self, other):
        if(isinstance(other, context.Context)):
            return other.__rmod__(self)
        elif(isinstance(other, basestring)):
            return repops_dim.DimRename(self, other)
        elif(isinstance(other, tuple)):
            return repops_dim.DimRename(self, *other)
        elif(isinstance(other, dict)):
            return repops_dim.DimRename(self, **other)
        return repops_funcs.Modulo(self, other)
    
    def __rmod__(self, other):
        return repops_funcs.Modulo(other, self)

    # division ( / )
    def __div__(self, other):
        if(isinstance(other, context.Context)):
            return other.__rdiv__(self)
        elif(isinstance(other, basestring)):
            return repops_slice.SliceRename(self, other)
        elif(isinstance(other, tuple)):
            return repops_slice.SliceRename(self, *other)
        elif(isinstance(other, dict)):
            return repops_slice.SliceRename(self, **other)
        return repops_funcs.Divide(self, other)
    
    def __rdiv__(self, other):
        return repops_funcs.Divide(other, self)
    
    def __floordiv__(self, other):
        if(isinstance(other, context.Context)):
            return other.__rfloordiv__(self)
        elif(isinstance(other, basestring)):
            return repops_slice.Bookmark(self, other)
        elif(isinstance(other, tuple)):
            return repops_slice.Bookmark(self, *other)
        elif(isinstance(other, dict)):
            return repops_slice.Bookmark(self, **other)
        return repops_funcs.FloorDivide(self, other)
    
    def __rfloordiv__(self,other):
        return repops_funcs.FloorDivide(other, self)
        

    def __pow__(self,other):
        if(isinstance(other, context.Context)):
            return other.__rpow__(self)
        return repops_funcs.Power(self, other)
    
    def __rpow__(self, other):
        return repops_funcs.Power(other,self)

    # and operator ( & )
    def __and__(self, other):
        if(isinstance(other, context.Context)):
            return other.__rand__(self)
        return repops_funcs.And(self, other)
    
    def __rand__(self, other):
        return repops_funcs.And(other, self)
    
    # or operator ( | )
    def __or__(self, other):
        if(isinstance(other, (context.Context, infix.Infix))):
            return other.__ror__(self)
        return repops_funcs.Or(self, other)

    def __ror__(self, other):
        return repops_funcs.Or(other, self)
    
    # exclusive-or operator ( ^ )
    def __xor__(self, other):
        if(isinstance(other, context.Context)):
            return other.__rxor__(self)
        return repops_funcs.Xor(self, other)

    def __rxor__(self, other):
        return repops_funcs.Xor(other, self)

    # less-than ( < )
    def __lt__(self, other):
        if(isinstance(other, context.Context)):
            return other.__gt__(self)
        return repops_funcs.Less(self, other)

    # less-than-or-equals ( <= )
    def __le__(self, other):
        if(isinstance(other, context.Context)):
            return other.__ge__(self)
        return repops_funcs.LessEqual(self, other)

    # equals ( == )
    def __eq__(self, other):
        if(isinstance(other, context.Context)):
            return other.__eq__(self)
        return repops_funcs.Equal(self, other)

    # not-equals ( != )
    def __ne__(self, other):
        if(isinstance(other, context.Context)):
            return other.__ne__(self)
        return repops_funcs.NotEqual(self, other)

    # greater-than ( > )
    def __gt__(self, other):
        if(isinstance(other, context.Context)):
            return other.__lt__(self)
        return repops_funcs.Greater(self, other)

    # greater-than-or-equals ( >= )
    def __ge__(self, other):
        if(isinstance(other, context.Context)):
            return other.__le__(self)
        return repops_funcs.GreaterEqual(self, other)

    # plus prefix, used for table operators (i.e. ++, &+, etc.)
    def __pos__(self):
        return repops.PlusPrefix(self)
    
    def __invert__(self):
        return repops_funcs.Invert(self)
       
    def __abs__(self):
        return repops_funcs.Abs(self)
    
    def __neg__(self):
        return repops_funcs.Negative(self)

    def __nonzero__(self):
        self._checkState()

        if(len(self._slices) != 1):
            raise RuntimeError, "Cannot determine True/False status of a multi-slice representor"

        slice = self._slices[0]
        if(len(slice.dims) > 1):
            nself = repops_dim.FlatAll(self)
        else:
            nself = self

        if(len(slice.dims) > 0):
            nself = repops_funcs.All(nself)
        
        res = nself()
        return bool(res)

    def Cast(self, *newtypes, **kwds):
        """Cast data to new type. 

        Allowed formats:

        * single type for all slices

          >>> x.Cast("int32")

        * type for each slice

          >>> x.Cast(("int64","int8"))

        * type for named slices

          >>> x.Cast(your_slice="int8")

        """
        return repops_slice.SliceCast(self, *newtypes, **kwds)

    def Transpose(self,permute_idxs=(1,0)):
        """Transposes the dimensions of slices.
           Can only be applied to the common dimensions of slices.

           :param permute_idxs: index order of new dims.
                By default, performs matrix transpose, of first two dims, 
                i.e. permute_idxs=(1,0)

        """
        return repops_dim.PermuteDims(self,permute_idxs)

    def Flat(self, dim=-1,name=None):
        """Flattens (merges) a dimension with previous(parent) dim. 

        :param dim: Dim to flatten. By default, last dim.
        :param name: Name of merged dim. By default, merged name of two dimensions.
        """
        return repops_dim.Flat(self, name=name,dim=dim)

    def FlatAll(self, name=None):
        """Flattens all dimensions into one dimension.

        :param name: name of new merged dimension. By default, merged names of all dimensions.
        """
        return repops_dim.FlatAll(self,name=name)

    def SplitDim(self,lshape,rshape,lname=None,rname=None,dimsel=None):
        """Splits dim into two dimensions.

        :param lshape: Left shape (integer or array of lengths)
        :param rshape: Right dim shape
        :param lname: New name of left dimension (default:autogenerated).
        :param rname: New name of right dimension (default:autogenerated).
        :param dimsel: Dim to split (default: last common dimension).
        """
        return repops_dim.SplitDim(self,lshape,rshape,lname,rname,dimsel)

    def Harray(self, name=None):
        """Combines slices into array.
        """
        return repops_slice.HArray(self, name=name)

    def getShape(self):
        """Returns shape of all dimensions as slices in a representor object"""
        return repops_dim.Shape(self)

    def GroupBy(self, *args, **kwargs):
        flat = kwargs.pop("flat", {})
        group_source = self.Get(*args, **kwargs)
        
        if(isinstance(flat,dict)):
            pass
        elif(isinstance(flat, (list,tuple))):
            flat = {0:flat}
        else:
            flat = {0:[flat]}
            
        return repops_multi.Group(self, group_source, flat)
   
    def Join(self, other, cond):
        return repops_multi.Join(self, other, cond)
    
    def Match(self, other, lslice=None,rslice=None, jointype="inner"):
        return repops_multi.Match(self, other, lslice, rslice, jointype)
    
    def Intersect(self, other, slices=COMMON_POS, dims=LASTCOMMONDIM, mode='dim'):
        return repops_multi.Intersect(self, other, slices=slices, dims=dims, mode=mode)
    
    def Except(self, other, slices=COMMON_POS, dims=LASTCOMMONDIM, mode='dim'):
        return repops_multi.Intersect(self, other, slices=slices, dims=dims, mode=mode)
    
    def Difference(self, other, slices=COMMON_POS, dims=LASTCOMMONDIM, mode='dim'):
        return repops_multi.Intersect(self, other, slices=slices, dims=dims, mode=mode)
   
    def Union(self, other, slices=COMMON_POS, dims=LASTCOMMONDIM, mode='dim'):
        return repops_multi.Intersect(self, other, slices=slices, dims=dims, mode=mode)

    def Replace(self, slice, translator, fromslice=0, toslice=1):
        return repops_multi.Replace(self, slice, translator, fromslice, toslice) 

    def ReplaceMissing(self, def_value=NOVAL):
        return repops_funcs.ReplaceMissing(self, def_value=def_value)

    def Merge(self, other):
        return repops_funcs.Merge(self, other)

    def Rename(self, *names, **kwds):
        """Rename slices.

        :param names: names without keywords. Number of names should match number of slices.
        :param kwds: names with keywords, e.g.  f0="genes". Number does *not* have to match number of slices.

        Examples::
            
            >>> na = a.Rename("genes","scores")
            >>> na = a.Rename(f0 = "genes", f1 = "scores")
        
        Shortcut:
            One can use the division operation to rename slices.

            * tuple:
                Number of names in tuple should match number of slices.

                >>> a/("genes", "scores")

            * dict:
                Similar to keywords.

                >>> a/{"f0": "genes"}

            * str:
                Can only be used if representor object consists of single slice.

                >>> a.f0/"genes"
        """ 
        return repops_slice.SliceRename(self, *names, **kwds)
    
    def DimRename(self, *names, **kwds):
        """Rename dimensions. Similar to ``Rename`` for slices.

        Shortcut: use % operator
        """
        return repops_dim.DimRename(self, *names, **kwds)
  
    def Bookmark(self, *names, **kwds):
        """Bookmarks slices with a name. 
        Slices can later be accessed using attribute access, 
        with axis indicator "B". 

        Example:
            >>> x = x.Bookmark("myslices")
            >>> x.myslices  
            >>> x.Bmyslices   #in case of possible conflicts with slice names
        """
        return repops_slice.Bookmark(self, *names, **kwds)

    def Each(self, eachfunc, dtype=rtypes.unknown):
        return repops_funcs.Each(self, eachfunc=eachfunc, dtype=dtype)
    
    def Pos(self, dim=None):
        return repops_funcs.Pos(self, dim)
    
    def CumSum(self, dim=None):
        return repops_funcs.CumSum(self, dim)
   
    def Argsort(self, dim=None, descend=False):
        return repops_funcs.Argsort(self, dim, descend=descend)
    
    def Rank(self, dim=None, descend=False):
        return repops_funcs.Rank(self, dim, descend=descend)
   
    def Argunique(self, dim=None):
        return repops_funcs.Argunique(self, dim)

    def Sum(self, dim=None):
        return repops_funcs.Sum(self, dim)

    def Max(self, dim=None):
        return repops_funcs.Max(self, dim)

    def Min(self, dim=None):
        return repops_funcs.Min(self, dim)
    
    def Argmax(self, dim=None):
        return repops_funcs.Argmax(self, dim)

    def Argmin(self, dim=None):
        return repops_funcs.Argmin(self, dim)

    def Mean(self, dim=None):
        return repops_funcs.Mean(self, dim)
    
    def Median(self, dim=None):
        return repops_funcs.Median(self, dim)
    
    def Std(self, dim=None):
        return repops_funcs.Std(self, dim)
     
    def Any(self,dim=None):
        return repops_funcs.Any(self,dim)

    def All(self,dim=None):
        return repops_funcs.All(self,dim)

    def Count(self):
        return repops_funcs.Count(self)

    def __len__(self):
        self._checkState()
        assert len(set([s.dims[0] for s in self._slices])) == 1, "To execute len, there should be only one root dimension"
        return self.Get(0).Array(1).Count()()
        
    def Set(self, dim=None):
        return repops_funcs.Set(self,dim)
    
    def Unique(self, dim=None):
        return repops_funcs.Unique(self,dim)

    def In(self, arrays):
        return repops_funcs.Within(self,arrays)

    def Contains(self, elems):
        return repops_funcs.Contains(self,elems)

    def Array(self, tolevel=None):
        """Packages dimension into array type"""
        return repops_dim.Array(self, tolevel = tolevel)

    def Level(self, tolevel):
        """Bring all slices to same dimension height throug packing and broadcasting"""
        return repops_dim.Level(self, tolevel)

    def Tuple(self):
        """Combines slices into a tuple type"""
        return repops_slice.Tuple(self)
    
    def Dict(self, with_missing=False):
        """Combines slices into a tuple type"""
        return repops_slice.Dict(self, with_missing=with_missing)
    
    def IndexDict(self):
        """Combines slices into a tuple type"""
        return repops_slice.IndexDict(self)

    def TakeFrom(self, other, allow_missing=False,keep_missing=False):
        return repops_multi.Take(other, self, allow_missing,keep_missing)

    def ToPython(self, **args):
        """Converts data into python data structure"""
        return repops_slice.ToPythonRep(self)(**args)

    def Sort(self, *slices, **kwargs):
        """Performs sort on data.
        
        Example:

        *  Sort slices in x on all slices. If multiple slices, combines into tuple, then sort it.

           >>> x.Sort()

        *  Sort x on slice f1

           >>> x.Sort(_.f1)

        * Sort x on slice f1, f3. 

          >>> x.Sort(_.f1, _.f3) 
        
        For other possible sort slice selection formats, see ``get`` function. 

        """
        descend = kwargs.pop("descend",False)

        if(slices or kwargs):
            sortsource = self.Get(*slices,**kwargs)
            return repops_multi.Sort(self, sortsource, descend=descend)
        else:
            return repops_multi.Sort(self, descend=descend)

    def Unique(self, *slices, **kwargs):
        if(slices or kwargs):
            sortsource = self.Get(*slices,**kwargs)
            return repops_multi.Unique(self, sortsource)
        else:
            return repops_multi.Unique(self)

    def To(self, *slices, **kwargs):
        return repops_slice.To(self, *slices, **kwargs)
              
              
    def AddSlice(self, name, data, dtype=None):
        return repops_slice.AddSlice(self, data, name, dtype)


    def Get(self, *slices, **kwds):
        """Select slices in a new representor, combine with other slices.

        :param slices: Can be various formats:

             * str:  selects slice with this name.
                     Special symbols:

                     - "*":  all slices

                     - "~": all slices with names not yet in previous ``get`` parameters are selected.

                     - "#": select first slice if all slices have a common dimension. 

             * int:   selects slice with this index

             * slice: selects slices with these indexes (note: a *Python* slice object, not an Ibidas slice)

             * representor: adds this representor slices

             * context: apply to self, adds resulting slices

             * tuple: applies get to elements in tuple, creates tuple slice from resulting slices (see .tuple() function)
             
             * list:  selected slices within list are packed using array function

        :param kwds: Same as slices, but also assigns slice names. 
        
        Examples:
            
            * str

              >>> a.Get("f0","f3")

            * int
    
              >>> a.Get(0, 3)

            * slice

              >>> a.Get(slice(0,3))
            
            * representor

              >>> a.Get(a.f0 + 3, a.f3 == "gene3",  Rep(3))

            * context

              >>> a.perform_some_operation().Get(_.f0 + 3, _.f3 == "gene3")
            
            * tuple

              >>> a.Get((_.f0, _.f1), _.f3)

            * list

              >>> a.Get([_.f0])

        """
        return repops_slice.Project(self,*slices,**kwds)

    def Without(self, *slices):
        return repops_slice.Unproject(self,*slices)

    def Elems(self, name=None):
        """Unpacks array type into dimension"""
        return repops_dim.UnpackArray(self, name)
    Elem = Elems
    def Fields(self, name=None):
        """Unpacks tuple type into slices"""
        return repops_slice.UnpackTuple(self, name)

    def GetRepeatedSliceNames(self):
        Pos = repops.delayable()(repops_funcs.Pos)
        Rep = python.Rep
        self._checkState()

        r = Rep(self.Names).Get(_/'name', Pos()/'pos').GroupBy(_.name)
        return r[_.pos.Count() > 1].Dict().ToPython()
        
        
            




import manager
import ensure_info
import create_graph
from ..utils import util
from ..utils.multi_visitor import VisitorFactory, NF_ERROR, F_CACHE
from ..constants import *
_delay_import_(globals(),"..representor")



class AnnotateRepLinks(VisitorFactory(prefixes=("link","distribute"), 
                                      flags=NF_ERROR | F_CACHE), manager.Pass):
    after=set([ensure_info.EnsureInfo,create_graph.CreateGraph])

    @classmethod
    def run(cls, query, run_manager):
        self = cls()
        query_root = query.root
        self.graph = run_manager.pass_results[create_graph.CreateGraph]
        self.links = self.graph.na['links']
        self.link(query_root)
        self.distribute(self.graph.root)
        return self.links

    def linkRepresentor(self,node):
        for slice in node._slices:
            if not slice in self.links and slice in self.graph.nodes:
                self.links[slice] = node

    def linkUnaryOpRep(self,node):
        self.link(node._source)
        self.linkRepresentor(node)
    
    def linkProject(self,node):
        self.link(node._source)
        if(node._req_sources):
            for name, elem in node._req_sources:
                if(isinstance(elem, representor.Representor)):
                    self.link(elem)
   
    def linkMultiOpRep(self,node):
        for source in node._sources:
            self.link(source)
        self.linkRepresentor(node)

    def distributeOp(self, node, lastlink=None):
        if(node in self.links):
            selflink = self.links[node]
        else:
            selflink = lastlink

        for edge in self.graph.edge_target[node]:
            sourcelink = self.distribute(edge.source, selflink)
                
        self.links[node] = selflink
        return selflink




from manager import Pass, PassManager
from ensure_info import EnsureInfo
from create_graph import CreateGraph
from annotate_replinks import AnnotateRepLinks
from prepeephole import PrePeepHole
from create_expression_objects import CreateExpressions
from wrapper_planner import WrapperPlanner
from python_peephole import PythonPeepHole
from transposer import Transposer
from serialize_exec import SerializeExec
from cytoscape_vis import DebugVisualizer




from collections import defaultdict
import random
import numpy

import manager
import create_graph
from ..utils.multi_visitor import VisitorFactory, DirectVisitorFactory, NF_ELSE
from ..utils import util
import xmlrpclib
from ibidas.utils.config import config
_delay_import_(globals(),"..representor")

networkid = util.seqgen().next
class DebugVisualizer(VisitorFactory(prefixes=("node",), flags=NF_ELSE),
                     DirectVisitorFactory(prefixes=("edge",), flags=NF_ELSE), manager.Pass):

    after = set([create_graph.CreateGraph])
    @classmethod
    def run(cls, query, run_manager, portnumber=config.get('debug.cytoscape_port_number',9000)):
        self = cls()
        self.rand = random.randint(0,10000000)
        self.graph = run_manager.pass_results[create_graph.CreateGraph]
        self.graph.pruneGraph()
        self.server = xmlrpclib.ServerProxy("http://localhost:" + str(portnumber)).Cytoscape
        self.network = self.server.createNetwork("network" + str(networkid()))
        
        self.unique_names = defaultdict(int)
        self.names = dict()

        self.node_name = dict()
        self.node_class = dict()
        self.node_type = dict()
        self.node_rep = dict()

        self.edge_from = []
        self.edge_to = []
        self.edge_type = []
        self.edge_attr = []

        for node in self.graph.nodes:
            self.node(node)

        self.server.createNodes(self.network, self.names.values())
        for source,edges in self.graph.edge_source.iteritems():
            for edge in edges:
                assert edge.source is source, "Source in edge and index not equal"
                self.edgeKey(edge.__class__.__name__,edge)
        
        self.edgeids = self.server.createEdges(self.network,self.edge_from, self.edge_to, self.edge_type,[True] * len(self.edge_type),False)

        self.server.addNodeAttributes("name","STRING",self.node_name,False)
        self.server.addNodeAttributes("type","STRING",self.node_type,False)
        self.server.addNodeAttributes("class","STRING",self.node_class,False)
        self.server.addNodeAttributes("rep","STRING",self.node_rep,False)
        self.server.addEdgeAttributes("type","STRING",dict(zip(self.edgeids,self.edge_type)))
        self.server.addEdgeAttributes("attr","STRING",dict(zip(self.edgeids,self.edge_attr)))

        for attribute,attribute_dict in self.graph.node_attributes.iteritems():
            attribute_name_dict = {}
            if(isinstance(attribute_dict.values()[0], float)):
                xtype = "FLOATING"
                cls = float
            elif(isinstance(attribute_dict.values()[0], int)):
                xtype = "INTEGER"
                cls = int
            else:
                xtype = "STRING"
                cls = str
            for node, node_name in self.names.iteritems():
                try:
                    r = attribute_dict.get(node,"")
                    if(isinstance(r, representor.Representor)):
                        r = str(r.__class__.__name__)
                    else:
                        r = cls(r)
                    attribute_name_dict[node_name] = r
                except:
                    pass
            self.server.addNodeAttributes(attribute,xtype,attribute_name_dict,False)
            if(attribute == 'links'):
                import matplotlib.cm
                cm = discrete_color_map(attribute_name_dict.values(), matplotlib.cm.gist_rainbow)
                self.server.createDiscreteMapper('default','links', 'Node Color','#444444',cm)
            if(attribute == "time"):
                self.server.createContinuousMapper('default','time', 'Node Size',[0.0, max(attribute_dict.values())],[20.0, 20.0, 100.0, 100.0])
    
            

        self.server.setNodeLabel(self.network, "name", "","default")
        self.server.setDiscreteNodeShapeMapper(self.network, 'default',
                'type', 'diamond', {'else':'ellipse', 'unaryop':'octagon', 'rep':'round_rect'}, True)
        self.server.setEdgeTargetArrowRule(self.network,"type","Arrow",["ParamListEdge","ParamChoiceListEdge"],["T","T"])
        self.server.setEdgeLineStyleRule(self.network,"type","SOLID",["ParamChoiceEdge","ParamChoiceListEdge",'SQLResultEdge'],["DOT","DOT","SINEWAVE"])
        self.server.performLayout(self.network, "hierarchical")
   
    def createUniqueName(self,name):
        counter = self.unique_names[name]
        self.unique_names[name] += 1
        counter -= 1

        if(counter < 0):
            return name + str(self.rand)
        else:
            return name + str(self.rand) + "_" + str(counter)

    def nodeelse(self,node):
        name = self.createUniqueName(node.__class__.__name__)
        self.names[node] = name
        if(node.__class__.__name__[-2:] == "Op"):
            self.node_name[name] = node.__class__.__name__[:-2]
        else:
            self.node_name[name] = node.__class__.__name__
        self.node_class[name] = node.__class__.__name__
        self.node_type[name] = "else"
        self.node_rep[name] = str(node)
        return name
    
    def nodeUnaryOp(self,node):
        name =self.nodeelse(node)
        self.node_type[name] = "unaryop"

    def edgeelse(self,edge):
        self.edge_from.append(self.names[edge.source])
        self.edge_to.append(self.names[edge.target])
        self.edge_type.append(edge.__class__.__name__)
        self.edge_attr.append(str(edge))




def discrete_color_map(elems, colormap):
    import matplotlib.colors
    cconv = matplotlib.colors.ColorConverter()
    z = set(elems)
    res = {}

    steps = numpy.linspace(0.0, 1.0, len(z))
    colors = colormap(steps)
    for elem, colorrow in zip(z, colors):
        res[elem] = matplotlib.colors.rgb2hex(cconv.to_rgb(colorrow))

    return res





from constants import *
import repops
from itertools import izip_longest

_delay_import_(globals(),"itypes","rtypes","dimpaths","dimensions")
_delay_import_(globals(),"ops")
_delay_import_(globals(),"repops_slice")
_delay_import_(globals(),"utils","util")

class UnpackArray(repops.UnaryOpRep):
    def _process(self,source, name = None, ndim=None):
        """Operation to unpack array typed slices

        :param source: source with active slices which should be unpacked
        :param name: (Optional) name of dimension to unpack. If not given,  unpack all.
        """
        if not source._typesKnown():
            return

        nslices = []                       #new active slices
        unpacked = False
        for slice in source._slices:
            #if name param, but does not match
            if(isinstance(slice.type,rtypes.TypeArray)):
                if(not name is None):
                    dimindex = slice.type.dims.getDimIndexByName(name)
                    if(not dimindex is None):
                        slice = ops.UnpackArrayOp(slice,ndim=dimindex+1)
                        unpacked = True
                else:
                    unpacked = True
                    slice = ops.UnpackArrayOp(slice,ndim=ndim)
            nslices.append(slice)
        if not unpacked:
            raise AttributeError, 'No array-typed slice with outer dimension name "' + name + '" to unpack'
        return self._initialize(tuple(nslices))

class DimRename(repops.UnaryOpRep):#{{{
    def _sprocess(self, source, *names, **kwds):
        if(names):
            unique_dimpath = util.unique(sum([slice.dims for slice in source._slices],dimpaths.DimPath())) 
            assert (len(names) <= len(unique_dimpath)),\
                    "Number of new dim names larger than number of dims"
            
            name_dict = {}
            for newname,dim  in zip(names,unique_dimpath):
                name_dict[dim] = newname
            kwds.update(name_dict)
        
        nslices = [ops.ChangeDimPathOp(slice, slice.dims.changeNames(kwds))
                                                    for slice in source._slices]
        return self._initialize(tuple(nslices))#}}}



class Redim(repops.UnaryOpRep):

    def _sprocess(self, source, *args, **kwds):
        dimname = args[0]
        dimsel = []
        slicesel = []

        for arg in args[1:]:
            if isinstance(arg, dict):
                kwds.update(arg)
            else:
                tslices = source.__getattr__(arg)._slices
                dimsel.extend([0] * len(tslices))
                slicesel.extend(tslices)
       
        for k,v in kwds.iteritems():
            tslices = source.__getattr__(k)._slices
            dimsel.extend([v] * len(tslices))
            slicesel.extend(tslices)
           
        nslices = list(source._slices)
        dims = set()
        for slice,dim in zip(slicesel,dimsel):
            slice = nslices[nslices.index(slice)]
            dimidxs = slice.dims.getDimIndices(dim)
            for dimidx in dimidxs:
                assert dimidx < len(slice.dims) and dimidx >= 0, 'Dim index should be >= 0 and < %d' % len(slice.dims)
                dims.add(slice.dims[dimidx])

        ndim = dimensions.toCommonDim(dimname, dims)
        for slice,dim in zip(slicesel,dimsel):
            sliceidx = nslices.index(slice)
            slice = nslices[sliceidx]
            dimidxs = slice.dims.getDimIndices(dim)
            for dimidx in dimidxs:
                slice = ops.ChangeDimOp(slice, dimidx, ndim)
            nslices[sliceidx] = slice                            
       
        return self._initialize(tuple(nslices))

class Shape(repops.UnaryOpRep):
    def _sprocess(self,source):
        dimnames = set()
        nslices = []
        for slice in source._slices:
            for pos, dim in enumerate(slice.dims):
                if dim.name in dimnames:
                    continue
                nslice = ops.ShapeOp(slice,pos)
                nslices.append(nslice)
                dimnames.add(dim.name)
        return self._initialize(tuple(nslices))
       

class InsertDim(repops.UnaryOpRep):
    def _sprocess(self, source, insertpoint, name=None):
        assert len(dimpaths.uniqueDimPath([s.dims for s in source._slices],only_unique=False)) >= insertpoint, "No unique dimpath covering dim insertion point"
        ndim = dimensions.Dim(1,name=name)
        nslices = []
        for slice in source._slices:
            slice = ops.InsertDimOp(slice,insertpoint,ndim)
            nslices.append(slice)
        return self._initialize(tuple(nslices))

class SplitDim(repops.UnaryOpRep):
    def _sprocess(self,source,lshape,rshape,lname=None,rname=None,dimsel=None):
        selpath = dimpaths.identifyUniqueDimPathSource(source,dimsel)
        nslices = []

        if(isinstance(lshape,int)):
            ldim = dimensions.Dim(lshape,name=lname)
        else:
            ldim = dimensions.Dim(UNDEFINED,dependent=(True,) * len(selpath.strip()), name=lname)

        if(isinstance(rshape,int)):
            rdim = dimensions.Dim(rshape,name=rname)
        else:
            rdim = dimensions.Dim(UNDEFINED,dependent=(True,) * (len(selpath.strip()) + 1), name=rname)
        
        for slice in source._slices:
            lastposs = slice.dims.matchDimPath(selpath)
            for lastpos in lastposs:
                slice = ops.SplitDimOp(slice,lastpos,lshape,rshape,ldim,rdim)
            nslices.append(slice)
            
        return self._initialize(tuple(nslices))


class PermuteDims(repops.UnaryOpRep):
    def _sprocess(self, source, permute_idxs):
        cpath = dimpaths.commonDimPath([s.dims for s in source._slices])
        if len(permute_idxs) > len(cpath):
            raise RuntimeError, "Permute index length longer than common dimensions"
        assert len(permute_idxs) == len(set(permute_idxs)), "Permute indices not unique"
        assert min(permute_idxs) == 0, "Lowest permute index should be 0"
        assert max(permute_idxs) == len(permute_idxs) - 1, "Highest permute index should be equal to: " + str(len(permute_idxs)-1)

        nslices = []
        for slice in source._slices:
            pidx = list(permute_idxs) + range(max(permute_idxs) + 1, len(slice.dims))
            slice = ops.PermuteDimsOp(slice,pidx)
            nslices.append(slice)
        return self._initialize(tuple(nslices))


class Array(repops.UnaryOpRep):
    def _sprocess(self, source, tolevel=None):
        if tolevel is None:
            nslices = [ops.PackArrayOp(slice) for slice in source._slices]
        else:
            nslices = []
            for slice in source._slices:
                if(len(slice.dims) > tolevel):
                    nslices.append(ops.PackArrayOp(slice, len(slice.dims) - tolevel))
                else:
                    nslices.append(slice)

        return self._initialize(tuple(nslices))


class Level(repops.UnaryOpRep):
    def _sprocess(self, source, tolevel=1):
        nslices = []
        for slice in source._slices:
            if(len(slice.dims) > tolevel):
                nslices.append(ops.PackArrayOp(slice, len(slice.dims) - tolevel))
            else:
                nslices.append(slice)

        if nslices:
            upaths = dimpaths.uniqueDimPath([s.dims for s in nslices],only_unique=True)
            assert len(upaths) >= tolevel, "Level can only be executed on slices with similar root dimensions"
        
        nslices,plan = ops.broadcast(nslices, mode="dim")
        return self._initialize(tuple(nslices))


class FlatAll(repops.UnaryOpRep):
    def _sprocess(self,source,name=None):
        nslices = ops.broadcast(source._slices,mode="dim")[0]
        dims = nslices[0].dims
        
        shape = 1
        for dim in dims:
            if(dim.shape == UNDEFINED):
                shape = UNDEFINED
                break
            else:
                shape *= dim.shape
        if(name is None):
            name = "_".join([dim.name for dim in dims])

        ndim = dimensions.Dim(shape, name=name)
        
        nnslices = []
        for slice in nslices:
            nnslices.append(ops.FlatAllOp(slice, ndim))

        return self._initialize(tuple(nnslices))

class Flat(repops.UnaryOpRep):
    def _sprocess(self,source,name=None,dim=-1):
        selpath = dimpaths.identifyUniqueDimPathSource(source,dim)
        nslices = self._apply(source._slices, selpath, name)

        return self._initialize(tuple(nslices))

    @classmethod
    def _apply(cls, fslices, selpath, name=None):
        #create new merged dimension
        selpath = dimpaths.extendParentDim(selpath,[s.dims for s in fslices],max(2,1 + len(selpath[-1].dependent)))
        
        #determine new dim
        if(selpath[-1].shape != UNDEFINED and selpath[-2].shape != UNDEFINED):
            shape = selpath[-2].shape * selpath[-1].shape
        else:
            shape = UNDEFINED

        if(name is None):
            name = selpath[-2].name + "_" + selpath[-1].name

        dependent = tuple([left or right for left,right in 
                        izip_longest(selpath[-2].dependent, selpath[-1].dependent[1:],fillvalue=False)])

        ndim = dimensions.Dim(shape, dependent=dependent, has_missing = selpath[-1].has_missing or selpath[-2].has_missing, name=name)
        
        bcdim = dimensions.Dim(1)

        #find refslices
        refslices = [s for s in fslices if selpath[-1] in s.dims]
        refslices2 = [s for s in fslices if selpath[-2] in s.dims]

        #process slices
        nslices = []

        for slice in fslices:
            sdims = slice.dims
            flatpos = sdims.matchDimPath(selpath[-1:])
            while flatpos:
                fpos = flatpos[-1]
                if(fpos == 0 or sdims[fpos-1] != selpath[-2]):
                    slice = ops.InsertDimOp(slice,fpos,bcdim)
                    bcdims = slice.dims[:fpos] + (selpath[-2],) + slice.dims[fpos:]
                    plan = [BCSOURCE] * len(slice.dims[:fpos]) + [BCEXIST] + [BCSOURCE] * len(slice.dims[fpos:])
                    slice = ops.BroadcastOp(slice,[refslices2],plan,bcdims)
                    fpos = fpos + 1
                
                slice = ops.FlatDimOp(slice,fpos,ndim)
                if(len(flatpos) > 1):
                    sdims = slice.dims
                    flatpos = sdims.matchDimPath(selpath[-1:])
                else:
                    flatpos = []
            sdims = slice.dims 
            lastpos = sdims.matchDimPath(selpath[-2:-1])
            while(lastpos):
                flatpos = lastpos[0] + 1
                if(len(sdims) <= flatpos or sdims[flatpos] != selpath[-1]):
                    slice = ops.InsertDimOp(slice,flatpos,bcdim)
                    bcdims = slice.dims[:flatpos] + (selpath[-1],) + slice.dims[flatpos:]
                    plan = [BCSOURCE] * len(slice.dims[:flatpos]) + [BCEXIST] + [BCSOURCE] * len(slice.dims[flatpos:])
                    slice = ops.BroadcastOp(slice,[refslices],plan,bcdims)
                slice = ops.FlatDimOp(slice,flatpos,ndim)
                if(len(lastpos) > 1):
                    sdims = slice.dims
                    lastpos = sdims.matchDimPath(selpath[-2:-1])
                else:
                    lastpos = []
                    
            nslices.append(slice)
        return nslices

class GroupIndex(repops.UnaryOpRep):
    def _sprocess(self, source):
        nslices = [ops.ensure_frozen(slice) for slice in source._slices]
        nslices = ops.broadcast(nslices,mode="dim")[0]
        nslices = [ops.PackArrayOp(nslice,1) for nslice in nslices]

        nslice = ops.GroupIndexOp(nslices)
        nslice = ops.UnpackArrayOp(nslice, len(nslices))
        return self._initialize((nslice,))



